#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Tests for 0D heat exchanger models.

Author: John Eslick
"""
from copy import deepcopy
import pytest
import pandas

from pyomo.environ import (
    check_optimal_termination,
    ConcreteModel,
    Constraint,
    Expression,
    value,
    Var,
    units as pyunits,
)
from pyomo.common.config import ConfigBlock

from idaes.core import (
    FlowsheetBlock,
    MaterialBalanceType,
    EnergyBalanceType,
    MomentumBalanceType,
)

from idaes.models.unit_models.heat_exchanger import (
    delta_temperature_lmtd_callback,
    delta_temperature_lmtd2_callback,
    delta_temperature_amtd_callback,
    delta_temperature_underwood_callback,
    delta_temperature_lmtd_smooth_callback,
    HeatExchanger,
    HeatExchangerFlowPattern,
    HX0DInitializer,
)
from idaes.models.properties.activity_coeff_models.BTX_activity_coeff_VLE import (
    BTXParameterBlock,
)
from idaes.models.properties import iapws95
from idaes.models.properties.examples.saponification_thermo import (
    SaponificationParameterBlock,
)
from idaes.models.properties.modular_properties.base.generic_property import (
    GenericParameterBlock,
)
from idaes.models.properties.modular_properties.examples.BT_PR import configuration
from idaes.core.util.model_statistics import (
    degrees_of_freedom,
    number_variables,
    number_total_constraints,
    number_unused_variables,
)
from idaes.core.util.testing import PhysicalParameterTestBlock, initialization_tester
from idaes.core.solvers import get_solver
from idaes.core.util.exceptions import InitializationError
from idaes.core.initialization import (
    BlockTriangularizationInitializer,
    InitializationStatus,
)
from idaes.core.util import DiagnosticsToolbox


# Imports to assemble BT-PR with different units
from idaes.core import LiquidPhase, VaporPhase, Component
from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.eos.ceos import Cubic, CubicType
from idaes.models.properties.modular_properties.phase_equil import SmoothVLE
from idaes.models.properties.modular_properties.phase_equil.bubble_dew import (
    LogBubbleDew,
)
from idaes.models.properties.modular_properties.phase_equil.forms import log_fugacity
import idaes.models.properties.modular_properties.pure.RPP4 as RPP
from idaes.core.util.exceptions import ConfigurationError
from idaes.models.properties.modular_properties.eos.ceos import cubic_roots_available

# -----------------------------------------------------------------------------
# Get default solver for testing
# TODO: Using MA57 causes a numerical diagnostics check to fail for some reason
# TODO: Instance of model is poorly scaled, so hopefully can be resolved with scaling tools
solver = get_solver(solver="ipopt_v2", solver_options={"linear_solver": "ma27"})


# -----------------------------------------------------------------------------
@pytest.mark.unit
def test_bad_option():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    with pytest.raises(KeyError):
        m.fs.unit = HeatExchanger(**{"I'm a bad option": "hot"})


@pytest.mark.unit
def test_bad_option2():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    with pytest.raises(
        ConfigurationError, match="cold_side_name cannot be 'hot_side'."
    ):
        m.fs.unit = HeatExchanger(cold_side_name="hot_side")


@pytest.mark.unit
def test_bad_option3():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    with pytest.raises(
        ConfigurationError, match="hot_side_name cannot be 'cold_side'."
    ):
        m.fs.unit = HeatExchanger(hot_side_name="cold_side")


@pytest.mark.unit
def test_bad_option4():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    with pytest.raises(
        ConfigurationError, match="cold_side_name cannot be 'cold_side'."
    ):
        m.fs.unit = HeatExchanger(cold_side_name="cold_side")


@pytest.mark.unit
def test_bad_option5():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    with pytest.raises(ConfigurationError, match="hot_side_name cannot be 'hot_side'."):
        m.fs.unit = HeatExchanger(hot_side_name="hot_side")


@pytest.mark.unit
def test_hot_and_cold_names_same():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    with pytest.raises(
        NameError,
        match="HeatExchanger hot and cold side cannot have the same name 'shell'.",
    ):
        m.fs.unit = HeatExchanger(hot_side_name="shell", cold_side_name="shell")


@pytest.mark.unit
def test_hot_side_name_clash():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = PhysicalParameterTestBlock()

    with pytest.raises(
        ValueError,
        match="fs.unit could not assign hot side alias "
        "build as an attribute of that name already "
        "exists.",
    ):
        m.fs.unit = HeatExchanger(
            hot_side={"property_package": m.fs.properties},
            cold_side={"property_package": m.fs.properties},
            hot_side_name="build",
        )


@pytest.mark.unit
def test_cold_side_name_clash():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = PhysicalParameterTestBlock()

    with pytest.raises(
        ValueError,
        match="fs.unit could not assign cold side alias "
        "build as an attribute of that name already "
        "exists.",
    ):
        m.fs.unit = HeatExchanger(
            hot_side={"property_package": m.fs.properties},
            cold_side={"property_package": m.fs.properties},
            cold_side_name="build",
        )


@pytest.mark.unit
def test_user_names():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)
    m.fs.properties = PhysicalParameterTestBlock()

    m.fs.unit = HeatExchanger(
        hot_side_name="shell",
        cold_side_name="tube",
        shell={"property_package": m.fs.properties},
        tube={"property_package": m.fs.properties},
    )

    assert m.fs.unit.config.hot_side.property_package is m.fs.properties
    assert m.fs.unit.config.cold_side.property_package is m.fs.properties

    assert m.fs.unit.shell is m.fs.unit.hot_side
    assert m.fs.unit.tube is m.fs.unit.cold_side

    assert m.fs.unit.shell_inlet is m.fs.unit.hot_side_inlet
    assert m.fs.unit.tube_inlet is m.fs.unit.cold_side_inlet
    assert m.fs.unit.shell_outlet is m.fs.unit.hot_side_outlet
    assert m.fs.unit.tube_outlet is m.fs.unit.cold_side_outlet


@pytest.mark.unit
def test_config():
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = PhysicalParameterTestBlock()

    m.fs.unit = HeatExchanger(
        hot_side={"property_package": m.fs.properties},
        cold_side={"property_package": m.fs.properties},
    )

    # Check unit config arguments
    # There are 8 to 10 arguments since you can add a side 1 and 2 config by
    # side_1, side_2, or whatever the user named them
    assert len(m.fs.unit.config) == 8

    assert not m.fs.unit.config.dynamic
    assert not m.fs.unit.config.has_holdup
    assert isinstance(m.fs.unit.config.hot_side, ConfigBlock)
    assert isinstance(m.fs.unit.config.cold_side, ConfigBlock)
    assert (
        m.fs.unit.config.delta_temperature_callback is delta_temperature_lmtd_callback
    )
    assert m.fs.unit.config.flow_pattern == HeatExchangerFlowPattern.countercurrent

    # Check hot_side config
    assert len(m.fs.unit.config.hot_side) == 7
    assert (
        m.fs.unit.config.hot_side.material_balance_type
        == MaterialBalanceType.useDefault
    )
    assert m.fs.unit.config.hot_side.energy_balance_type == EnergyBalanceType.useDefault
    assert (
        m.fs.unit.config.hot_side.momentum_balance_type
        == MomentumBalanceType.pressureTotal
    )
    assert not m.fs.unit.config.hot_side.has_phase_equilibrium
    assert not m.fs.unit.config.hot_side.has_pressure_change
    assert m.fs.unit.config.hot_side.property_package is m.fs.properties

    # Check cold_side config
    assert len(m.fs.unit.config.cold_side) == 7
    assert (
        m.fs.unit.config.cold_side.material_balance_type
        == MaterialBalanceType.useDefault
    )
    assert (
        m.fs.unit.config.cold_side.energy_balance_type == EnergyBalanceType.useDefault
    )
    assert (
        m.fs.unit.config.cold_side.momentum_balance_type
        == MomentumBalanceType.pressureTotal
    )
    assert not m.fs.unit.config.cold_side.has_phase_equilibrium
    assert not m.fs.unit.config.cold_side.has_pressure_change
    assert m.fs.unit.config.cold_side.property_package is m.fs.properties

    assert m.fs.unit.default_initializer is HX0DInitializer


def basic_model(cb=delta_temperature_lmtd_callback):
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = iapws95.Iapws95ParameterBlock()

    m.fs.unit = HeatExchanger(
        hot_side={"property_package": m.fs.properties},
        cold_side={"property_package": m.fs.properties},
        delta_temperature_callback=cb,
        flow_pattern=HeatExchangerFlowPattern.countercurrent,
    )
    #   Set inputs
    m.fs.unit.hot_side_inlet.flow_mol[0].fix(100)
    m.fs.unit.hot_side_inlet.enth_mol[0].fix(4000)
    m.fs.unit.hot_side_inlet.pressure[0].fix(101325)

    m.fs.unit.cold_side_inlet.flow_mol[0].fix(100)
    m.fs.unit.cold_side_inlet.enth_mol[0].fix(3500)
    m.fs.unit.cold_side_inlet.pressure[0].fix(101325)

    m.fs.unit.area.fix(1000)
    m.fs.unit.overall_heat_transfer_coefficient.fix(100)

    assert degrees_of_freedom(m) == 0
    m.fs.unit.initialize()
    return m


def basic_model2(cb=delta_temperature_lmtd_callback):
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = iapws95.Iapws95ParameterBlock()

    m.fs.unit = HeatExchanger(
        hot_side={"property_package": m.fs.properties},
        cold_side={"property_package": m.fs.properties},
        delta_temperature_callback=cb,
        flow_pattern=HeatExchangerFlowPattern.cocurrent,
    )
    #   Set inputs
    m.fs.unit.hot_side_inlet.flow_mol[0].fix(100)
    m.fs.unit.hot_side_inlet.enth_mol[0].fix(4000)
    m.fs.unit.hot_side_inlet.pressure[0].fix(101325)

    m.fs.unit.cold_side_inlet.flow_mol[0].fix(100)
    m.fs.unit.cold_side_inlet.enth_mol[0].fix(3500)
    m.fs.unit.cold_side_inlet.pressure[0].fix(101325)

    m.fs.unit.area.fix(100)
    m.fs.unit.overall_heat_transfer_coefficient.fix(100)

    assert degrees_of_freedom(m) == 0
    m.fs.unit.initialize()
    return m


def basic_model3(cb=delta_temperature_lmtd_callback):
    m = ConcreteModel()
    m.fs = FlowsheetBlock(dynamic=False)

    m.fs.properties = iapws95.Iapws95ParameterBlock()

    m.fs.unit = HeatExchanger(
        hot_side={"property_package": m.fs.properties},
        cold_side={"property_package": m.fs.properties},
        delta_temperature_callback=cb,
        flow_pattern=HeatExchangerFlowPattern.crossflow,
    )
    #   Set inputs
    m.fs.unit.hot_side_inlet.flow_mol[0].fix(100)
    m.fs.unit.hot_side_inlet.enth_mol[0].fix(4000)
    m.fs.unit.hot_side_inlet.pressure[0].fix(101325)

    m.fs.unit.cold_side_inlet.flow_mol[0].fix(100)
    m.fs.unit.cold_side_inlet.enth_mol[0].fix(3500)
    m.fs.unit.cold_side_inlet.pressure[0].fix(101325)

    m.fs.unit.area.fix(1000)
    m.fs.unit.overall_heat_transfer_coefficient.fix(100)
    m.fs.unit.crossflow_factor.fix(1.0)
    assert degrees_of_freedom(m) == 0
    m.fs.unit.initialize()
    return m


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_lmtd_smooth_cb():
    m = basic_model(delta_temperature_lmtd_smooth_callback)
    results = solver.solve(m)
    assert check_optimal_termination(results)
    # hot in end
    assert value(m.fs.unit.delta_temperature_in[0]) == pytest.approx(0.464879, rel=1e-3)
    # hot out end
    assert value(m.fs.unit.delta_temperature_out[0]) == pytest.approx(
        0.465069, rel=1e-3
    )
    assert value(m.fs.unit.heat_duty[0]) == pytest.approx(46497.44)


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_lmtd2_cb():
    m = basic_model(delta_temperature_lmtd2_callback)
    results = solver.solve(m)
    assert check_optimal_termination(results)
    # hot in end
    assert value(m.fs.unit.delta_temperature_in[0]) == pytest.approx(0.464879, rel=1e-3)
    # hot out end
    assert value(m.fs.unit.delta_temperature_out[0]) == pytest.approx(
        0.465069, rel=1e-3
    )
    assert value(m.fs.unit.heat_duty[0]) == pytest.approx(46497.44)


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_lmtd2_co_cb():
    m = basic_model2(delta_temperature_lmtd2_callback)
    results = solver.solve(m)
    assert check_optimal_termination(results)
    # hot in end
    assert value(m.fs.unit.delta_temperature_in[0]) == pytest.approx(6.63771, rel=1e-3)
    # hot out end
    assert value(m.fs.unit.delta_temperature_out[0]) == pytest.approx(0.46658, rel=1e-3)
    assert value(m.fs.unit.heat_duty[0]) == pytest.approx(23242.69)


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_lmtd2_cross_cb():
    m = basic_model3(delta_temperature_lmtd2_callback)
    results = solver.solve(m)
    assert check_optimal_termination(results)
    # hot in end
    assert value(m.fs.unit.delta_temperature_in[0]) == pytest.approx(0.464879, rel=1e-3)
    # hot out end
    assert value(m.fs.unit.delta_temperature_out[0]) == pytest.approx(
        0.465069, rel=1e-3
    )
    assert value(m.fs.unit.heat_duty[0]) == pytest.approx(46497.44)


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_lmtd3_cb():
    m = basic_model(delta_temperature_lmtd2_callback)
    results = solver.solve(m)
    assert check_optimal_termination(results)
    # hot in end
    assert value(m.fs.unit.delta_temperature_in[0]) == pytest.approx(0.464879, rel=1e-3)
    # hot out end
    assert value(m.fs.unit.delta_temperature_out[0]) == pytest.approx(
        0.465069, rel=1e-3
    )
    assert value(m.fs.unit.heat_duty[0]) == pytest.approx(46497.44)


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_amtd_cb():
    # since the delta T at both ends, AMTD ends up about the same as LMTD
    m = basic_model(delta_temperature_amtd_callback)
    results = solver.solve(m)
    assert check_optimal_termination(results)
    # hot in end
    assert value(m.fs.unit.delta_temperature_in[0]) == pytest.approx(0.464879, rel=1e-3)
    # hot out end
    assert value(m.fs.unit.delta_temperature_out[0]) == pytest.approx(
        0.465069, rel=1e-3
    )
    assert value(m.fs.unit.heat_duty[0]) == pytest.approx(46497.44)


@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
@pytest.mark.skipif(solver is None, reason="Solver not available")
@pytest.mark.component
def test_underwood_cb():
    m = basic_model(delta_temperature_underwood_callback)
    results = solver.solve(m)
    assert check_optimal_termination(results)
    # hot in end
    assert value(m.fs.unit.delta_temperature_in[0]) == pytest.approx(0.464879, rel=1e-3)
    # hot out end
    assert value(m.fs.unit.delta_temperature_out[0]) == pytest.approx(
        0.465069, rel=1e-3
    )
    assert value(m.fs.unit.heat_duty[0]) == pytest.approx(46497.44)


# -----------------------------------------------------------------------------
class TestBTX_cocurrent(object):
    @pytest.fixture(scope="class")
    def btx(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = BTXParameterBlock(valid_phase="Liq")

        m.fs.unit = HeatExchanger(
            hot_side={"property_package": m.fs.properties},
            cold_side={"property_package": m.fs.properties},
            flow_pattern=HeatExchangerFlowPattern.cocurrent,
        )

        m.fs.unit.hot_side_inlet.flow_mol[0].fix(5)  # mol/s
        m.fs.unit.hot_side_inlet.temperature[0].fix(365)  # K
        m.fs.unit.hot_side_inlet.pressure[0].fix(101325)  # Pa
        m.fs.unit.hot_side_inlet.mole_frac_comp[0, "benzene"].fix(0.5)
        m.fs.unit.hot_side_inlet.mole_frac_comp[0, "toluene"].fix(0.5)

        m.fs.unit.cold_side_inlet.flow_mol[0].fix(1)  # mol/s
        m.fs.unit.cold_side_inlet.temperature[0].fix(300)  # K
        m.fs.unit.cold_side_inlet.pressure[0].fix(101325)  # Pa
        m.fs.unit.cold_side_inlet.mole_frac_comp[0, "benzene"].fix(0.5)
        m.fs.unit.cold_side_inlet.mole_frac_comp[0, "toluene"].fix(0.5)

        m.fs.unit.area.fix(1)
        m.fs.unit.overall_heat_transfer_coefficient.fix(100)

        # Bound temperature differences to avoid division by zero
        m.fs.unit.delta_temperature_in[0.0].setlb(40)
        m.fs.unit.delta_temperature_out[0.0].setlb(0.1)

        m.fs.unit.delta_temperature_in[0.0].setub(80)
        m.fs.unit.delta_temperature_out[0.0].setub(35)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, btx):
        assert hasattr(btx.fs.unit, "hot_side_inlet")
        assert len(btx.fs.unit.hot_side_inlet.vars) == 4
        assert hasattr(btx.fs.unit.hot_side_inlet, "flow_mol")
        assert hasattr(btx.fs.unit.hot_side_inlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.hot_side_inlet, "temperature")
        assert hasattr(btx.fs.unit.hot_side_inlet, "pressure")

        assert hasattr(btx.fs.unit, "cold_side_inlet")
        assert len(btx.fs.unit.cold_side_inlet.vars) == 4
        assert hasattr(btx.fs.unit.cold_side_inlet, "flow_mol")
        assert hasattr(btx.fs.unit.cold_side_inlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.cold_side_inlet, "temperature")
        assert hasattr(btx.fs.unit.cold_side_inlet, "pressure")

        assert hasattr(btx.fs.unit, "hot_side_outlet")
        assert len(btx.fs.unit.hot_side_outlet.vars) == 4
        assert hasattr(btx.fs.unit.hot_side_outlet, "flow_mol")
        assert hasattr(btx.fs.unit.hot_side_outlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.hot_side_outlet, "temperature")
        assert hasattr(btx.fs.unit.hot_side_outlet, "pressure")

        assert hasattr(btx.fs.unit, "cold_side_outlet")
        assert len(btx.fs.unit.cold_side_outlet.vars) == 4
        assert hasattr(btx.fs.unit.cold_side_outlet, "flow_mol")
        assert hasattr(btx.fs.unit.cold_side_outlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.cold_side_outlet, "temperature")
        assert hasattr(btx.fs.unit.cold_side_outlet, "pressure")

        assert isinstance(btx.fs.unit.overall_heat_transfer_coefficient, Var)
        assert isinstance(btx.fs.unit.area, Var)
        assert not hasattr(btx.fs.unit, "crossflow_factor")
        assert isinstance(btx.fs.unit.heat_duty, Var)
        assert isinstance(btx.fs.unit.delta_temperature_in, Var)
        assert isinstance(btx.fs.unit.delta_temperature_out, Var)
        assert isinstance(btx.fs.unit.unit_heat_balance, Constraint)
        assert isinstance(btx.fs.unit.delta_temperature, (Var, Expression))
        assert isinstance(btx.fs.unit.heat_transfer_equation, Constraint)

        assert number_variables(btx) == 50
        assert number_total_constraints(btx) == 38
        assert number_unused_variables(btx) == 0

    @pytest.mark.component
    def test_structural_issues(self, btx):
        dt = DiagnosticsToolbox(btx)
        dt.display_potential_evaluation_errors()
        dt.assert_no_structural_warnings()

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_performance_contents(self, btx):
        perf_dict = btx.fs.unit._get_performance_contents()

        assert perf_dict == {
            "vars": {
                "HX Area": btx.fs.unit.area,
                "Heat Duty": btx.fs.unit.heat_duty[0],
                "HX Coefficient": btx.fs.unit.overall_heat_transfer_coefficient[0],
                "Delta T In": btx.fs.unit.delta_temperature_in[0],
                "Delta T Out": btx.fs.unit.delta_temperature_out[0],
            },
            "exprs": {
                "Delta T Driving": btx.fs.unit.delta_temperature[0],
            },
        }

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_stream_table_contents(self, btx):
        stable = btx.fs.unit._get_stream_table_contents()

        expected = pandas.DataFrame.from_dict(
            {
                "Units": {
                    "flow_mol": getattr(pyunits.pint_registry, "mole/second"),
                    "mole_frac_comp benzene": getattr(
                        pyunits.pint_registry, "dimensionless"
                    ),
                    "mole_frac_comp toluene": getattr(
                        pyunits.pint_registry, "dimensionless"
                    ),
                    "temperature": getattr(pyunits.pint_registry, "kelvin"),
                    "pressure": getattr(pyunits.pint_registry, "Pa"),
                },
                "Hot Side Inlet": {
                    "flow_mol": 5.0,
                    "mole_frac_comp benzene": 0.5,
                    "mole_frac_comp toluene": 0.5,
                    "temperature": 365,
                    "pressure": 101325.0,
                },
                "Hot Side Outlet": {
                    "flow_mol": 1.0,
                    "mole_frac_comp benzene": 0.5,
                    "mole_frac_comp toluene": 0.5,
                    "temperature": 298.15,
                    "pressure": 101325.0,
                },
                "Cold Side Inlet": {
                    "flow_mol": 1.0,
                    "mole_frac_comp benzene": 0.5,
                    "mole_frac_comp toluene": 0.5,
                    "temperature": 300,
                    "pressure": 101325.0,
                },
                "Cold Side Outlet": {
                    "flow_mol": 1.0,
                    "mole_frac_comp benzene": 0.5,
                    "mole_frac_comp toluene": 0.5,
                    "temperature": 298.15,
                    "pressure": 101325.0,
                },
            }
        )

        pandas.testing.assert_frame_equal(stable, expected, rtol=1e-4, atol=1e-4)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, btx):
        initialization_tester(btx)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, btx):
        results = solver.solve(btx)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, btx):
        assert pytest.approx(5, abs=1e-3) == value(
            btx.fs.unit.hot_side_outlet.flow_mol[0]
        )
        assert pytest.approx(359.5, abs=1e-1) == value(
            btx.fs.unit.hot_side_outlet.temperature[0]
        )
        assert pytest.approx(101325, abs=1e-3) == value(
            btx.fs.unit.hot_side_outlet.pressure[0]
        )

        assert pytest.approx(1, abs=1e-3) == value(
            btx.fs.unit.cold_side_outlet.flow_mol[0]
        )
        assert pytest.approx(329.9, abs=1e-1) == value(
            btx.fs.unit.cold_side_outlet.temperature[0]
        )
        assert pytest.approx(101325, abs=1e-3) == value(
            btx.fs.unit.cold_side_outlet.pressure[0]
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, btx):
        assert (
            abs(
                value(
                    btx.fs.unit.hot_side_inlet.flow_mol[0]
                    - btx.fs.unit.hot_side_outlet.flow_mol[0]
                )
            )
            <= 1e-6
        )
        assert (
            abs(
                value(
                    btx.fs.unit.cold_side_inlet.flow_mol[0]
                    - btx.fs.unit.cold_side_outlet.flow_mol[0]
                )
            )
            <= 1e-6
        )

        hot_side = value(
            btx.fs.unit.hot_side_outlet.flow_mol[0]
            * (
                btx.fs.unit.hot_side.properties_in[0].enth_mol_phase["Liq"]
                - btx.fs.unit.hot_side.properties_out[0].enth_mol_phase["Liq"]
            )
        )
        cold_side = value(
            btx.fs.unit.cold_side_outlet.flow_mol[0]
            * (
                btx.fs.unit.cold_side.properties_in[0].enth_mol_phase["Liq"]
                - btx.fs.unit.cold_side.properties_out[0].enth_mol_phase["Liq"]
            )
        )
        assert abs(hot_side + cold_side) <= 1e-6

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_numerical_issues(self, btx):
        dt = DiagnosticsToolbox(btx)
        dt.assert_no_numerical_warnings()


# -----------------------------------------------------------------------------
class TestBTX_cocurrent_alt_name(object):
    @pytest.fixture(scope="class")
    def btx(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = BTXParameterBlock(valid_phase="Liq")

        m.fs.unit = HeatExchanger(
            hot_side_name="hot",
            cold_side_name="cold",
            hot={"property_package": m.fs.properties},
            cold={"property_package": m.fs.properties},
            flow_pattern=HeatExchangerFlowPattern.cocurrent,
        )

        m.fs.unit.hot_inlet.flow_mol[0].fix(5)  # mol/s
        m.fs.unit.hot_inlet.temperature[0].fix(365)  # K
        m.fs.unit.hot_inlet.pressure[0].fix(101325)  # Pa
        m.fs.unit.hot_inlet.mole_frac_comp[0, "benzene"].fix(0.5)
        m.fs.unit.hot_inlet.mole_frac_comp[0, "toluene"].fix(0.5)

        m.fs.unit.cold_inlet.flow_mol[0].fix(1)  # mol/s
        m.fs.unit.cold_inlet.temperature[0].fix(300)  # K
        m.fs.unit.cold_inlet.pressure[0].fix(101325)  # Pa
        m.fs.unit.cold_inlet.mole_frac_comp[0, "benzene"].fix(0.5)
        m.fs.unit.cold_inlet.mole_frac_comp[0, "toluene"].fix(0.5)

        m.fs.unit.area.fix(1)
        m.fs.unit.overall_heat_transfer_coefficient.fix(100)

        # Bound temperature differences to avoid division by zero
        m.fs.unit.delta_temperature_in[0.0].setlb(40)
        m.fs.unit.delta_temperature_out[0.0].setlb(0.1)

        m.fs.unit.delta_temperature_in[0.0].setub(80)
        m.fs.unit.delta_temperature_out[0.0].setub(35)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, btx):
        assert hasattr(btx.fs.unit, "hot_inlet")
        assert len(btx.fs.unit.hot_inlet.vars) == 4
        assert hasattr(btx.fs.unit.hot_inlet, "flow_mol")
        assert hasattr(btx.fs.unit.hot_inlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.hot_inlet, "temperature")
        assert hasattr(btx.fs.unit.hot_inlet, "pressure")

        assert hasattr(btx.fs.unit, "cold_inlet")
        assert len(btx.fs.unit.cold_inlet.vars) == 4
        assert hasattr(btx.fs.unit.cold_inlet, "flow_mol")
        assert hasattr(btx.fs.unit.cold_inlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.cold_inlet, "temperature")
        assert hasattr(btx.fs.unit.cold_inlet, "pressure")

        assert hasattr(btx.fs.unit, "hot_outlet")
        assert len(btx.fs.unit.hot_outlet.vars) == 4
        assert hasattr(btx.fs.unit.hot_outlet, "flow_mol")
        assert hasattr(btx.fs.unit.hot_outlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.hot_outlet, "temperature")
        assert hasattr(btx.fs.unit.hot_outlet, "pressure")

        assert hasattr(btx.fs.unit, "cold_outlet")
        assert len(btx.fs.unit.cold_outlet.vars) == 4
        assert hasattr(btx.fs.unit.cold_outlet, "flow_mol")
        assert hasattr(btx.fs.unit.cold_outlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.cold_outlet, "temperature")
        assert hasattr(btx.fs.unit.cold_outlet, "pressure")

        assert isinstance(btx.fs.unit.overall_heat_transfer_coefficient, Var)
        assert isinstance(btx.fs.unit.area, Var)
        assert not hasattr(btx.fs.unit, "crossflow_factor")
        assert isinstance(btx.fs.unit.heat_duty, Var)
        assert isinstance(btx.fs.unit.delta_temperature_in, Var)
        assert isinstance(btx.fs.unit.delta_temperature_out, Var)
        assert isinstance(btx.fs.unit.unit_heat_balance, Constraint)
        assert isinstance(btx.fs.unit.delta_temperature, (Var, Expression))
        assert isinstance(btx.fs.unit.heat_transfer_equation, Constraint)

        assert number_variables(btx) == 50
        assert number_total_constraints(btx) == 38
        assert number_unused_variables(btx) == 0

    @pytest.mark.component
    def test_structural_issues(self, btx):
        dt = DiagnosticsToolbox(btx)
        dt.assert_no_structural_warnings()

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_performance_contents(self, btx):
        perf_dict = btx.fs.unit._get_performance_contents()

        assert perf_dict == {
            "vars": {
                "HX Area": btx.fs.unit.area,
                "Heat Duty": btx.fs.unit.heat_duty[0],
                "HX Coefficient": btx.fs.unit.overall_heat_transfer_coefficient[0],
                "Delta T In": btx.fs.unit.delta_temperature_in[0],
                "Delta T Out": btx.fs.unit.delta_temperature_out[0],
            },
            "exprs": {
                "Delta T Driving": btx.fs.unit.delta_temperature[0],
            },
        }

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_stream_table_contents(self, btx):
        stable = btx.fs.unit._get_stream_table_contents()

        expected = pandas.DataFrame.from_dict(
            {
                "Units": {
                    "flow_mol": getattr(pyunits.pint_registry, "mole/second"),
                    "mole_frac_comp benzene": getattr(
                        pyunits.pint_registry, "dimensionless"
                    ),
                    "mole_frac_comp toluene": getattr(
                        pyunits.pint_registry, "dimensionless"
                    ),
                    "temperature": getattr(pyunits.pint_registry, "kelvin"),
                    "pressure": getattr(pyunits.pint_registry, "Pa"),
                },
                "hot Inlet": {
                    "flow_mol": 5.0,
                    "mole_frac_comp benzene": 0.5,
                    "mole_frac_comp toluene": 0.5,
                    "temperature": 365,
                    "pressure": 101325.0,
                },
                "hot Outlet": {
                    "flow_mol": 1.0,
                    "mole_frac_comp benzene": 0.5,
                    "mole_frac_comp toluene": 0.5,
                    "temperature": 298.15,
                    "pressure": 101325.0,
                },
                "cold Inlet": {
                    "flow_mol": 1.0,
                    "mole_frac_comp benzene": 0.5,
                    "mole_frac_comp toluene": 0.5,
                    "temperature": 300,
                    "pressure": 101325.0,
                },
                "cold Outlet": {
                    "flow_mol": 1.0,
                    "mole_frac_comp benzene": 0.5,
                    "mole_frac_comp toluene": 0.5,
                    "temperature": 298.15,
                    "pressure": 101325.0,
                },
            }
        )

        pandas.testing.assert_frame_equal(stable, expected, rtol=1e-4, atol=1e-4)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, btx):
        initialization_tester(btx)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, btx):
        results = solver.solve(btx)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, btx):
        assert pytest.approx(5, abs=1e-3) == value(btx.fs.unit.hot_outlet.flow_mol[0])
        assert pytest.approx(359.5, abs=1e-1) == value(
            btx.fs.unit.hot_outlet.temperature[0]
        )
        assert pytest.approx(101325, abs=1e-3) == value(
            btx.fs.unit.hot_outlet.pressure[0]
        )

        assert pytest.approx(1, abs=1e-3) == value(btx.fs.unit.cold_outlet.flow_mol[0])
        assert pytest.approx(329.9, abs=1e-1) == value(
            btx.fs.unit.cold_outlet.temperature[0]
        )
        assert pytest.approx(101325, abs=1e-3) == value(
            btx.fs.unit.cold_outlet.pressure[0]
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, btx):
        assert (
            abs(
                value(
                    btx.fs.unit.hot_inlet.flow_mol[0]
                    - btx.fs.unit.hot_outlet.flow_mol[0]
                )
            )
            <= 1e-6
        )
        assert (
            abs(
                value(
                    btx.fs.unit.cold_inlet.flow_mol[0]
                    - btx.fs.unit.cold_outlet.flow_mol[0]
                )
            )
            <= 1e-6
        )

        hot_side = value(
            btx.fs.unit.hot_outlet.flow_mol[0]
            * (
                btx.fs.unit.hot.properties_in[0].enth_mol_phase["Liq"]
                - btx.fs.unit.hot.properties_out[0].enth_mol_phase["Liq"]
            )
        )
        cold_side = value(
            btx.fs.unit.cold_outlet.flow_mol[0]
            * (
                btx.fs.unit.cold.properties_in[0].enth_mol_phase["Liq"]
                - btx.fs.unit.cold.properties_out[0].enth_mol_phase["Liq"]
            )
        )
        assert abs(hot_side + cold_side) <= 1e-6

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_numerical_issues(self, btx):
        dt = DiagnosticsToolbox(btx)
        dt.assert_no_numerical_warnings()


# -----------------------------------------------------------------------------
@pytest.mark.iapws
@pytest.mark.skipif(not iapws95.iapws95_available(), reason="IAPWS not available")
class TestIAPWS_countercurrent(object):
    @pytest.fixture(scope="class")
    def iapws(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = iapws95.Iapws95ParameterBlock()

        m.fs.unit = HeatExchanger(
            hot_side={"property_package": m.fs.properties},
            cold_side={"property_package": m.fs.properties},
            flow_pattern=HeatExchangerFlowPattern.countercurrent,
        )

        m.fs.unit.hot_side_inlet.flow_mol[0].fix(100)
        m.fs.unit.hot_side_inlet.enth_mol[0].fix(6000)
        m.fs.unit.hot_side_inlet.pressure[0].fix(101325)

        m.fs.unit.cold_side_inlet.flow_mol[0].fix(100)
        m.fs.unit.cold_side_inlet.enth_mol[0].fix(5000)
        m.fs.unit.cold_side_inlet.pressure[0].fix(101325)

        m.fs.unit.area.fix(1000)
        m.fs.unit.overall_heat_transfer_coefficient.fix(100)

        return m

    @pytest.fixture(scope="class")
    def iapws_underwood(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = iapws95.Iapws95ParameterBlock()

        m.fs.unit = HeatExchanger(
            hot_side={"property_package": m.fs.properties},
            cold_side={"property_package": m.fs.properties},
            delta_temperature_callback=delta_temperature_underwood_callback,
            flow_pattern=HeatExchangerFlowPattern.countercurrent,
        )

        m.fs.unit.hot_side_inlet.flow_mol[0].fix(100)
        m.fs.unit.hot_side_inlet.enth_mol[0].fix(6000)
        m.fs.unit.hot_side_inlet.pressure[0].fix(101325)

        m.fs.unit.cold_side_inlet.flow_mol[0].fix(100)
        m.fs.unit.cold_side_inlet.enth_mol[0].fix(5000)
        m.fs.unit.cold_side_inlet.pressure[0].fix(101325)

        m.fs.unit.area.fix(1000)
        m.fs.unit.overall_heat_transfer_coefficient.fix(100)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, iapws):
        assert len(iapws.fs.unit.hot_side_inlet.vars) == 3
        assert hasattr(iapws.fs.unit.hot_side_inlet, "flow_mol")
        assert hasattr(iapws.fs.unit.hot_side_inlet, "enth_mol")
        assert hasattr(iapws.fs.unit.hot_side_inlet, "pressure")

        assert hasattr(iapws.fs.unit, "hot_side_outlet")
        assert len(iapws.fs.unit.hot_side_outlet.vars) == 3
        assert hasattr(iapws.fs.unit.hot_side_outlet, "flow_mol")
        assert hasattr(iapws.fs.unit.hot_side_outlet, "enth_mol")
        assert hasattr(iapws.fs.unit.hot_side_outlet, "pressure")

        assert len(iapws.fs.unit.cold_side_inlet.vars) == 3
        assert hasattr(iapws.fs.unit.cold_side_inlet, "flow_mol")
        assert hasattr(iapws.fs.unit.cold_side_inlet, "enth_mol")
        assert hasattr(iapws.fs.unit.cold_side_inlet, "pressure")

        assert hasattr(iapws.fs.unit, "cold_side_outlet")
        assert len(iapws.fs.unit.cold_side_outlet.vars) == 3
        assert hasattr(iapws.fs.unit.cold_side_outlet, "flow_mol")
        assert hasattr(iapws.fs.unit.cold_side_outlet, "enth_mol")
        assert hasattr(iapws.fs.unit.cold_side_outlet, "pressure")

        assert isinstance(iapws.fs.unit.overall_heat_transfer_coefficient, Var)
        assert isinstance(iapws.fs.unit.area, Var)
        assert not hasattr(iapws.fs.unit, "crossflow_factor")
        assert isinstance(iapws.fs.unit.heat_duty, Var)
        assert isinstance(iapws.fs.unit.delta_temperature_in, Var)
        assert isinstance(iapws.fs.unit.delta_temperature_out, Var)
        assert isinstance(iapws.fs.unit.unit_heat_balance, Constraint)
        assert isinstance(iapws.fs.unit.delta_temperature, (Expression, Var))
        assert isinstance(iapws.fs.unit.heat_transfer_equation, Constraint)

        assert number_variables(iapws) == 18
        assert number_total_constraints(iapws) == 10
        assert number_unused_variables(iapws) == 0

    @pytest.mark.component
    def test_structural_issues(self, iapws):
        dt = DiagnosticsToolbox(iapws)
        # Delta T calculations cause potential evaluation errors that are hard to bound
        dt.assert_no_structural_warnings(ignore_evaluation_errors=True)

    @pytest.mark.unit
    def test_dof_alt_name1(self, iapws):
        iapws.fs.unit.hot_side_inlet.flow_mol[0].fix(100)
        iapws.fs.unit.hot_side_inlet.enth_mol[0].fix(6000)
        iapws.fs.unit.hot_side_inlet.pressure[0].fix(101325)

        iapws.fs.unit.cold_side_inlet.flow_mol[0].fix(100)
        iapws.fs.unit.cold_side_inlet.enth_mol[0].fix(5000)
        iapws.fs.unit.cold_side_inlet.pressure[0].fix(101325)

        iapws.fs.unit.area.fix(1000)
        iapws.fs.unit.overall_heat_transfer_coefficient.fix(100)

        assert degrees_of_freedom(iapws) == 0

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_performance_contents(self, iapws):
        perf_dict = iapws.fs.unit._get_performance_contents()

        assert perf_dict == {
            "vars": {
                "HX Area": iapws.fs.unit.area,
                "Heat Duty": iapws.fs.unit.heat_duty[0],
                "HX Coefficient": iapws.fs.unit.overall_heat_transfer_coefficient[0],
                "Delta T In": iapws.fs.unit.delta_temperature_in[0],
                "Delta T Out": iapws.fs.unit.delta_temperature_out[0],
            },
            "exprs": {
                "Delta T Driving": iapws.fs.unit.delta_temperature[0],
            },
        }

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_stream_table_contents(self, iapws):
        stable = iapws.fs.unit._get_stream_table_contents()

        expected = pandas.DataFrame.from_dict(
            {
                "Units": {
                    "Molar Flow": getattr(pyunits.pint_registry, "mole/second"),
                    "Mass Flow": getattr(pyunits.pint_registry, "kg/second"),
                    "T": getattr(pyunits.pint_registry, "K"),
                    "P": getattr(pyunits.pint_registry, "Pa"),
                    "Vapor Fraction": getattr(pyunits.pint_registry, "dimensionless"),
                    "Molar Enthalpy": getattr(pyunits.pint_registry, "J/mole"),
                },
                "Hot Side Inlet": {
                    "Molar Flow": 100,
                    "Mass Flow": 1.8015,
                    "T": 352.67,
                    "P": 101325,
                    "Vapor Fraction": 0,
                    "Molar Enthalpy": 6000.0,
                },
                "Hot Side Outlet": {
                    "Molar Flow": 1,
                    "Mass Flow": 1.8015e-2,
                    "T": 270.4877112932641,
                    "P": 11032305.8275,
                    "Vapor Fraction": 0,
                    "Molar Enthalpy": 0.01102138712926277,
                },
                "Cold Side Inlet": {
                    "Molar Flow": 100,
                    "Mass Flow": 1.8015,
                    "T": 339.43,
                    "P": 101325,
                    "Vapor Fraction": 0,
                    "Molar Enthalpy": 5000.0,
                },
                "Cold Side Outlet": {
                    "Molar Flow": 1,
                    "Mass Flow": 1.8015e-2,
                    "T": 270.4877112932641,
                    "P": 11032305.8275,
                    "Vapor Fraction": 0,
                    "Molar Enthalpy": 0.01102138712926277,
                },
            }
        )

        pandas.testing.assert_frame_equal(stable, expected, rtol=1e-4, atol=1e-4)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, iapws):
        initialization_tester(iapws)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, iapws):
        results = solver.solve(iapws)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize_underwood(self, iapws_underwood):
        initialization_tester(iapws_underwood)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve_underwood(self, iapws_underwood):
        results = solver.solve(iapws_underwood)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, iapws):
        assert pytest.approx(100, abs=1e-5) == value(
            iapws.fs.unit.hot_side_outlet.flow_mol[0]
        )
        assert pytest.approx(100, abs=1e-5) == value(
            iapws.fs.unit.cold_side_outlet.flow_mol[0]
        )

        assert pytest.approx(5070.219, abs=1e0) == value(
            iapws.fs.unit.hot_side_outlet.enth_mol[0]
        )
        assert pytest.approx(5929.781, abs=1e0) == value(
            iapws.fs.unit.cold_side_outlet.enth_mol[0]
        )

        assert pytest.approx(101325, abs=1e2) == value(
            iapws.fs.unit.hot_side_outlet.pressure[0]
        )
        assert pytest.approx(101325, abs=1e2) == value(
            iapws.fs.unit.cold_side_outlet.pressure[0]
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, iapws):
        assert (
            abs(
                value(
                    iapws.fs.unit.hot_side_inlet.flow_mol[0]
                    - iapws.fs.unit.hot_side_outlet.flow_mol[0]
                )
            )
            <= 1e-6
        )
        assert (
            abs(
                value(
                    iapws.fs.unit.cold_side_inlet.flow_mol[0]
                    - iapws.fs.unit.cold_side_outlet.flow_mol[0]
                )
            )
            <= 1e-6
        )

        hot_side = value(
            iapws.fs.unit.hot_side_outlet.flow_mol[0]
            * (
                iapws.fs.unit.hot_side_inlet.enth_mol[0]
                - iapws.fs.unit.hot_side_outlet.enth_mol[0]
            )
        )
        cold_side = value(
            iapws.fs.unit.cold_side_outlet.flow_mol[0]
            * (
                iapws.fs.unit.cold_side_inlet.enth_mol[0]
                - iapws.fs.unit.cold_side_outlet.enth_mol[0]
            )
        )
        assert abs(hot_side + cold_side) <= 1e-6

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_numerical_issues(self, iapws):
        dt = DiagnosticsToolbox(iapws)
        dt.assert_no_numerical_warnings()


# -----------------------------------------------------------------------------
class TestSaponification_crossflow(object):
    @pytest.fixture(scope="class")
    def sapon(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = SaponificationParameterBlock()

        m.fs.unit = HeatExchanger(
            hot_side={"property_package": m.fs.properties},
            cold_side={"property_package": m.fs.properties},
            flow_pattern=HeatExchangerFlowPattern.crossflow,
        )

        m.fs.unit.hot_side_inlet.flow_vol[0].fix(1e-3)
        m.fs.unit.hot_side_inlet.temperature[0].fix(320)
        m.fs.unit.hot_side_inlet.pressure[0].fix(101325)
        m.fs.unit.hot_side_inlet.conc_mol_comp[0, "H2O"].fix(55388.0)
        m.fs.unit.hot_side_inlet.conc_mol_comp[0, "NaOH"].fix(100.0)
        m.fs.unit.hot_side_inlet.conc_mol_comp[0, "EthylAcetate"].fix(100.0)
        m.fs.unit.hot_side_inlet.conc_mol_comp[0, "SodiumAcetate"].fix(1e-12)
        m.fs.unit.hot_side_inlet.conc_mol_comp[0, "Ethanol"].fix(1e-12)

        m.fs.unit.cold_side_inlet.flow_vol[0].fix(1e-3)
        m.fs.unit.cold_side_inlet.temperature[0].fix(300)
        m.fs.unit.cold_side_inlet.pressure[0].fix(101325)
        m.fs.unit.cold_side_inlet.conc_mol_comp[0, "H2O"].fix(55388.0)
        m.fs.unit.cold_side_inlet.conc_mol_comp[0, "NaOH"].fix(100.0)
        m.fs.unit.cold_side_inlet.conc_mol_comp[0, "EthylAcetate"].fix(100.0)
        m.fs.unit.cold_side_inlet.conc_mol_comp[0, "SodiumAcetate"].fix(1e-12)
        m.fs.unit.cold_side_inlet.conc_mol_comp[0, "Ethanol"].fix(1e-12)

        m.fs.unit.area.fix(1000)
        m.fs.unit.overall_heat_transfer_coefficient.fix(100)
        m.fs.unit.crossflow_factor.fix(0.6)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, sapon):
        assert len(sapon.fs.unit.hot_side_inlet.vars) == 4
        assert hasattr(sapon.fs.unit.hot_side_inlet, "flow_vol")
        assert hasattr(sapon.fs.unit.hot_side_inlet, "conc_mol_comp")
        assert hasattr(sapon.fs.unit.hot_side_inlet, "temperature")
        assert hasattr(sapon.fs.unit.hot_side_inlet, "pressure")

        assert len(sapon.fs.unit.hot_side_outlet.vars) == 4
        assert hasattr(sapon.fs.unit.hot_side_outlet, "flow_vol")
        assert hasattr(sapon.fs.unit.hot_side_outlet, "conc_mol_comp")
        assert hasattr(sapon.fs.unit.hot_side_outlet, "temperature")
        assert hasattr(sapon.fs.unit.hot_side_outlet, "pressure")

        assert len(sapon.fs.unit.cold_side_inlet.vars) == 4
        assert hasattr(sapon.fs.unit.cold_side_inlet, "flow_vol")
        assert hasattr(sapon.fs.unit.cold_side_inlet, "conc_mol_comp")
        assert hasattr(sapon.fs.unit.cold_side_inlet, "temperature")
        assert hasattr(sapon.fs.unit.cold_side_inlet, "pressure")

        assert len(sapon.fs.unit.cold_side_outlet.vars) == 4
        assert hasattr(sapon.fs.unit.cold_side_outlet, "flow_vol")
        assert hasattr(sapon.fs.unit.cold_side_outlet, "conc_mol_comp")
        assert hasattr(sapon.fs.unit.cold_side_outlet, "temperature")
        assert hasattr(sapon.fs.unit.cold_side_outlet, "pressure")

        assert isinstance(sapon.fs.unit.overall_heat_transfer_coefficient, Var)
        assert isinstance(sapon.fs.unit.area, Var)
        assert isinstance(sapon.fs.unit.crossflow_factor, Var)
        assert isinstance(sapon.fs.unit.heat_duty, Var)
        assert isinstance(sapon.fs.unit.delta_temperature_in, Var)
        assert isinstance(sapon.fs.unit.delta_temperature_out, Var)
        assert isinstance(sapon.fs.unit.unit_heat_balance, Constraint)
        assert isinstance(sapon.fs.unit.delta_temperature, (Expression, Var))
        assert isinstance(sapon.fs.unit.heat_transfer_equation, Constraint)

        assert number_variables(sapon) == 39
        assert number_total_constraints(sapon) == 20
        assert number_unused_variables(sapon) == 0

    @pytest.mark.component
    def test_structural_issues(self, sapon):
        dt = DiagnosticsToolbox(sapon)
        # Delta T calculations cause potential evaluation errors that are hard to bound
        dt.assert_no_structural_warnings(ignore_evaluation_errors=True)

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_performance_contents(self, sapon):
        perf_dict = sapon.fs.unit._get_performance_contents()

        assert perf_dict == {
            "vars": {
                "HX Area": sapon.fs.unit.area,
                "Heat Duty": sapon.fs.unit.heat_duty[0],
                "HX Coefficient": sapon.fs.unit.overall_heat_transfer_coefficient[0],
                "Crossflow Factor": sapon.fs.unit.crossflow_factor[0],
                "Delta T In": sapon.fs.unit.delta_temperature_in[0],
                "Delta T Out": sapon.fs.unit.delta_temperature_out[0],
            },
            "exprs": {
                "Delta T Driving": sapon.fs.unit.delta_temperature[0],
            },
        }

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_stream_table_contents(self, sapon):
        stable = sapon.fs.unit._get_stream_table_contents()

        expected = pandas.DataFrame.from_dict(
            {
                "Units": {
                    "Volumetric Flowrate": getattr(
                        pyunits.pint_registry, "m**3/second"
                    ),
                    "Molar Concentration H2O": getattr(
                        pyunits.pint_registry, "mole/m**3"
                    ),
                    "Molar Concentration NaOH": getattr(
                        pyunits.pint_registry, "mole/m**3"
                    ),
                    "Molar Concentration EthylAcetate": getattr(
                        pyunits.pint_registry, "mole/m**3"
                    ),
                    "Molar Concentration SodiumAcetate": getattr(
                        pyunits.pint_registry, "mole/m**3"
                    ),
                    "Molar Concentration Ethanol": getattr(
                        pyunits.pint_registry, "mole/m**3"
                    ),
                    "Temperature": getattr(pyunits.pint_registry, "K"),
                    "Pressure": getattr(pyunits.pint_registry, "Pa"),
                },
                "Hot Side Inlet": {
                    "Volumetric Flowrate": 1e-3,
                    "Molar Concentration H2O": 55388,
                    "Molar Concentration NaOH": 100.00,
                    "Molar Concentration EthylAcetate": 100.00,
                    "Molar Concentration SodiumAcetate": 0,
                    "Molar Concentration Ethanol": 0,
                    "Temperature": 320,
                    "Pressure": 1.0132e05,
                },
                "Hot Side Outlet": {
                    "Volumetric Flowrate": 1.00,
                    "Molar Concentration H2O": 100.00,
                    "Molar Concentration NaOH": 100.00,
                    "Molar Concentration EthylAcetate": 100.00,
                    "Molar Concentration SodiumAcetate": 100.00,
                    "Molar Concentration Ethanol": 100.00,
                    "Temperature": 298.15,
                    "Pressure": 1.0132e05,
                },
                "Cold Side Inlet": {
                    "Volumetric Flowrate": 1e-3,
                    "Molar Concentration H2O": 55388,
                    "Molar Concentration NaOH": 100.00,
                    "Molar Concentration EthylAcetate": 100.00,
                    "Molar Concentration SodiumAcetate": 0,
                    "Molar Concentration Ethanol": 0,
                    "Temperature": 300,
                    "Pressure": 1.0132e05,
                },
                "Cold Side Outlet": {
                    "Volumetric Flowrate": 1.00,
                    "Molar Concentration H2O": 100.00,
                    "Molar Concentration NaOH": 100.00,
                    "Molar Concentration EthylAcetate": 100.00,
                    "Molar Concentration SodiumAcetate": 100.00,
                    "Molar Concentration Ethanol": 100.00,
                    "Temperature": 298.15,
                    "Pressure": 1.0132e05,
                },
            }
        )

        pandas.testing.assert_frame_equal(stable, expected, rtol=1e-4, atol=1e-4)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, sapon):
        initialization_tester(sapon)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, sapon):
        results = solver.solve(sapon)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, sapon):
        assert pytest.approx(1e-3, abs=1e-6) == value(
            sapon.fs.unit.hot_side_outlet.flow_vol[0]
        )
        assert pytest.approx(1e-3, abs=1e-6) == value(
            sapon.fs.unit.cold_side_outlet.flow_vol[0]
        )

        assert pytest.approx(55388.0, rel=1e-3) == value(
            sapon.fs.unit.hot_side_outlet.conc_mol_comp[0, "H2O"]
        )
        assert pytest.approx(100.0, rel=1e-3) == value(
            sapon.fs.unit.hot_side_outlet.conc_mol_comp[0, "NaOH"]
        )
        assert pytest.approx(100.0, rel=1e-3) == value(
            sapon.fs.unit.hot_side_outlet.conc_mol_comp[0, "EthylAcetate"]
        )
        assert pytest.approx(0.0, abs=1e-3) == value(
            sapon.fs.unit.hot_side_outlet.conc_mol_comp[0, "SodiumAcetate"]
        )
        assert pytest.approx(0.0, abs=1e-3) == value(
            sapon.fs.unit.hot_side_outlet.conc_mol_comp[0, "Ethanol"]
        )

        assert pytest.approx(55388.0, rel=1e-3) == value(
            sapon.fs.unit.cold_side_outlet.conc_mol_comp[0, "H2O"]
        )
        assert pytest.approx(100.0, rel=1e-3) == value(
            sapon.fs.unit.cold_side_outlet.conc_mol_comp[0, "NaOH"]
        )
        assert pytest.approx(100.0, rel=1e-3) == value(
            sapon.fs.unit.cold_side_outlet.conc_mol_comp[0, "EthylAcetate"]
        )
        assert pytest.approx(0.0, abs=1e-3) == value(
            sapon.fs.unit.cold_side_outlet.conc_mol_comp[0, "SodiumAcetate"]
        )
        assert pytest.approx(0.0, abs=1e-3) == value(
            sapon.fs.unit.cold_side_outlet.conc_mol_comp[0, "Ethanol"]
        )

        assert pytest.approx(301.3, abs=1e-1) == value(
            sapon.fs.unit.hot_side_outlet.temperature[0]
        )
        assert pytest.approx(318.7, abs=1e-1) == value(
            sapon.fs.unit.cold_side_outlet.temperature[0]
        )

        assert pytest.approx(101325, abs=1e2) == value(
            sapon.fs.unit.hot_side_outlet.pressure[0]
        )
        assert pytest.approx(101325, abs=1e2) == value(
            sapon.fs.unit.cold_side_outlet.pressure[0]
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, sapon):
        hot_side = value(
            sapon.fs.unit.hot_side_outlet.flow_vol[0]
            * sapon.fs.properties.dens_mol
            * sapon.fs.properties.cp_mol
            * (
                sapon.fs.unit.hot_side_inlet.temperature[0]
                - sapon.fs.unit.hot_side_outlet.temperature[0]
            )
        )
        cold_side = value(
            sapon.fs.unit.cold_side_outlet.flow_vol[0]
            * sapon.fs.properties.dens_mol
            * sapon.fs.properties.cp_mol
            * (
                sapon.fs.unit.cold_side_inlet.temperature[0]
                - sapon.fs.unit.cold_side_outlet.temperature[0]
            )
        )
        assert abs(hot_side + cold_side) <= 1e0

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_numerical_issues(self, sapon):
        # TODO: Heat transfer constraint has a residual of ~1e-3
        # Model could be better scaled
        # TODO: Using MA57 results in extreme Jacobian and aprallel constraints?
        dt = DiagnosticsToolbox(sapon, constraint_residual_tolerance=1e-2)
        dt.assert_no_numerical_warnings()


# -----------------------------------------------------------------------------
@pytest.mark.skipif(not cubic_roots_available(), reason="Cubic functions not available")
class TestBT_Generic_cocurrent(object):
    @pytest.fixture(scope="class")
    def btx(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        # As we lack other example prop packs with units, take the generic
        # BT-PR package and change the base units
        configuration2 = {
            # Specifying components
            "components": {
                "benzene": {
                    "type": Component,
                    "enth_mol_ig_comp": RPP,
                    "entr_mol_ig_comp": RPP,
                    "pressure_sat_comp": RPP,
                    "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
                    "parameter_data": {
                        "mw": (78.1136e-3, pyunits.kg / pyunits.mol),  # [1]
                        "pressure_crit": (48.9e5, pyunits.Pa),  # [1]
                        "temperature_crit": (562.2, pyunits.K),  # [1]
                        "omega": 0.212,  # [1]
                        "cp_mol_ig_comp_coeff": {
                            "A": (-3.392e1, pyunits.J / pyunits.mol / pyunits.K),  # [1]
                            "B": (4.739e-1, pyunits.J / pyunits.mol / pyunits.K**2),
                            "C": (-3.017e-4, pyunits.J / pyunits.mol / pyunits.K**3),
                            "D": (7.130e-8, pyunits.J / pyunits.mol / pyunits.K**4),
                        },
                        "enth_mol_form_vap_comp_ref": (
                            82.9e3,
                            pyunits.J / pyunits.mol,
                        ),  # [3]
                        "entr_mol_form_vap_comp_ref": (
                            -269,
                            pyunits.J / pyunits.mol / pyunits.K,
                        ),  # [3]
                        "pressure_sat_comp_coeff": {
                            "A": (-6.98273, None),  # [1]
                            "B": (1.33213, None),
                            "C": (-2.62863, None),
                            "D": (-3.33399, None),
                        },
                    },
                },
                "toluene": {
                    "type": Component,
                    "enth_mol_ig_comp": RPP,
                    "entr_mol_ig_comp": RPP,
                    "pressure_sat_comp": RPP,
                    "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
                    "parameter_data": {
                        "mw": (92.1405e-3, pyunits.kg / pyunits.mol),  # [1]
                        "pressure_crit": (41e5, pyunits.Pa),  # [1]
                        "temperature_crit": (591.8, pyunits.K),  # [1]
                        "omega": 0.263,  # [1]
                        "cp_mol_ig_comp_coeff": {
                            "A": (-2.435e1, pyunits.J / pyunits.mol / pyunits.K),  # [1]
                            "B": (5.125e-1, pyunits.J / pyunits.mol / pyunits.K**2),
                            "C": (-2.765e-4, pyunits.J / pyunits.mol / pyunits.K**3),
                            "D": (4.911e-8, pyunits.J / pyunits.mol / pyunits.K**4),
                        },
                        "enth_mol_form_vap_comp_ref": (
                            50.1e3,
                            pyunits.J / pyunits.mol,
                        ),  # [3]
                        "entr_mol_form_vap_comp_ref": (
                            -321,
                            pyunits.J / pyunits.mol / pyunits.K,
                        ),  # [3]
                        "pressure_sat_comp_coeff": {
                            "A": (-7.28607, None),  # [1]
                            "B": (1.38091, None),
                            "C": (-2.83433, None),
                            "D": (-2.79168, None),
                        },
                    },
                },
            },
            # Specifying phases
            "phases": {
                "Liq": {
                    "type": LiquidPhase,
                    "equation_of_state": Cubic,
                    "equation_of_state_options": {"type": CubicType.PR},
                },
                "Vap": {
                    "type": VaporPhase,
                    "equation_of_state": Cubic,
                    "equation_of_state_options": {"type": CubicType.PR},
                },
            },
            # Set base units of measurement
            "base_units": {
                "time": pyunits.s,
                "length": pyunits.m,
                "mass": pyunits.t,
                "amount": pyunits.mol,
                "temperature": pyunits.degR,
            },
            # Specifying state definition
            "state_definition": FTPx,
            "state_bounds": {
                "flow_mol": (0, 100, 1000, pyunits.mol / pyunits.s),
                "temperature": (273.15, 300, 500, pyunits.K),
                "pressure": (5e4, 1e5, 1e6, pyunits.Pa),
            },
            "pressure_ref": (101325, pyunits.Pa),
            "temperature_ref": (298.15, pyunits.K),
            # Defining phase equilibria
            "phases_in_equilibrium": [("Vap", "Liq")],
            "phase_equilibrium_state": {("Vap", "Liq"): SmoothVLE},
            "bubble_dew_method": LogBubbleDew,
            "parameter_data": {
                "PR_kappa": {
                    ("benzene", "benzene"): 0.000,
                    ("benzene", "toluene"): 0.000,
                    ("toluene", "benzene"): 0.000,
                    ("toluene", "toluene"): 0.000,
                }
            },
        }

        m.fs.properties = GenericParameterBlock(**configuration)
        m.fs.properties2 = GenericParameterBlock(**configuration2)

        m.fs.unit = HeatExchanger(
            hot_side={"property_package": m.fs.properties},
            cold_side={"property_package": m.fs.properties2},
            flow_pattern=HeatExchangerFlowPattern.cocurrent,
        )

        m.fs.unit.hot_side_inlet.flow_mol[0].fix(5)  # mol/s
        m.fs.unit.hot_side_inlet.temperature[0].fix(365)  # K
        m.fs.unit.hot_side_inlet.pressure[0].fix(101325)  # Pa
        m.fs.unit.hot_side_inlet.mole_frac_comp[0, "benzene"].fix(0.5)
        m.fs.unit.hot_side_inlet.mole_frac_comp[0, "toluene"].fix(0.5)

        m.fs.unit.cold_side_inlet.flow_mol[0].fix(1)  # mol/s
        m.fs.unit.cold_side_inlet.temperature[0].fix(540)  # degR
        m.fs.unit.cold_side_inlet.pressure[0].fix(101.325)  # kPa
        m.fs.unit.cold_side_inlet.mole_frac_comp[0, "benzene"].fix(0.5)
        m.fs.unit.cold_side_inlet.mole_frac_comp[0, "toluene"].fix(0.5)

        m.fs.unit.area.fix(1)
        m.fs.unit.overall_heat_transfer_coefficient.fix(100)

        m.fs.unit.cold_side.scaling_factor_pressure = 1

        # Set small values of epsilon to get sufficiently accurate results
        # Only applies to hot side, as cold side used the original SmoothVLE.
        m.fs.unit.hot_side.properties_in[0].eps_t_Vap_Liq.set_value(1e-4)
        m.fs.unit.hot_side.properties_in[0].eps_z_Vap_Liq.set_value(1e-4)
        m.fs.unit.hot_side.properties_out[0].eps_t_Vap_Liq.set_value(1e-4)
        m.fs.unit.hot_side.properties_out[0].eps_z_Vap_Liq.set_value(1e-4)

        return m

    @pytest.mark.build
    @pytest.mark.unit
    def test_build(self, btx):
        assert hasattr(btx.fs.unit, "hot_side_inlet")
        assert len(btx.fs.unit.hot_side_inlet.vars) == 4
        assert hasattr(btx.fs.unit.hot_side_inlet, "flow_mol")
        assert hasattr(btx.fs.unit.hot_side_inlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.hot_side_inlet, "temperature")
        assert hasattr(btx.fs.unit.hot_side_inlet, "pressure")

        assert hasattr(btx.fs.unit, "cold_side_inlet")
        assert len(btx.fs.unit.cold_side_inlet.vars) == 4
        assert hasattr(btx.fs.unit.cold_side_inlet, "flow_mol")
        assert hasattr(btx.fs.unit.cold_side_inlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.cold_side_inlet, "temperature")
        assert hasattr(btx.fs.unit.cold_side_inlet, "pressure")

        assert hasattr(btx.fs.unit, "hot_side_outlet")
        assert len(btx.fs.unit.hot_side_outlet.vars) == 4
        assert hasattr(btx.fs.unit.hot_side_outlet, "flow_mol")
        assert hasattr(btx.fs.unit.hot_side_outlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.hot_side_outlet, "temperature")
        assert hasattr(btx.fs.unit.hot_side_outlet, "pressure")

        assert hasattr(btx.fs.unit, "cold_side_outlet")
        assert len(btx.fs.unit.cold_side_outlet.vars) == 4
        assert hasattr(btx.fs.unit.cold_side_outlet, "flow_mol")
        assert hasattr(btx.fs.unit.cold_side_outlet, "mole_frac_comp")
        assert hasattr(btx.fs.unit.cold_side_outlet, "temperature")
        assert hasattr(btx.fs.unit.cold_side_outlet, "pressure")

        assert isinstance(btx.fs.unit.overall_heat_transfer_coefficient, Var)
        assert isinstance(btx.fs.unit.area, Var)
        assert not hasattr(btx.fs.unit, "crossflow_factor")
        assert isinstance(btx.fs.unit.heat_duty, Var)
        assert isinstance(btx.fs.unit.delta_temperature_in, Var)
        assert isinstance(btx.fs.unit.delta_temperature_out, Var)
        assert isinstance(btx.fs.unit.unit_heat_balance, Constraint)
        assert isinstance(btx.fs.unit.delta_temperature, (Var, Expression))
        assert isinstance(btx.fs.unit.heat_transfer_equation, Constraint)

        assert number_variables(btx) == 176
        assert number_total_constraints(btx) == 104
        assert number_unused_variables(btx) == 20

    @pytest.mark.component
    def test_structural_issues(self, btx):
        dt = DiagnosticsToolbox(btx)
        # Delta T calculations cause potential evaluation errors that are hard to bound
        dt.assert_no_structural_warnings(ignore_evaluation_errors=True)

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_performance_contents(self, btx):
        perf_dict = btx.fs.unit._get_performance_contents()

        assert perf_dict == {
            "vars": {
                "HX Area": btx.fs.unit.area,
                "Heat Duty": btx.fs.unit.heat_duty[0],
                "HX Coefficient": btx.fs.unit.overall_heat_transfer_coefficient[0],
                "Delta T In": btx.fs.unit.delta_temperature_in[0],
                "Delta T Out": btx.fs.unit.delta_temperature_out[0],
            },
            "exprs": {
                "Delta T Driving": btx.fs.unit.delta_temperature[0],
            },
        }

    @pytest.mark.ui
    @pytest.mark.unit
    def test_get_stream_table_contents(self, btx):
        stable = btx.fs.unit._get_stream_table_contents()

        expected = pandas.DataFrame.from_dict(
            {
                "Units": {
                    "Total Molar Flowrate": getattr(
                        pyunits.pint_registry, "mole/second"
                    ),
                    "Total Mole Fraction benzene": getattr(
                        pyunits.pint_registry, "dimensionless"
                    ),
                    "Total Mole Fraction toluene": getattr(
                        pyunits.pint_registry, "dimensionless"
                    ),
                    "Temperature": getattr(pyunits.pint_registry, "kelvin"),
                    "Pressure": getattr(pyunits.pint_registry, "Pa"),
                },
                "Hot Side Inlet": {
                    "Total Molar Flowrate": 5.0,
                    "Total Mole Fraction benzene": 0.5,
                    "Total Mole Fraction toluene": 0.5,
                    "Temperature": 365,
                    "Pressure": 101325.0,
                },
                "Hot Side Outlet": {
                    "Total Molar Flowrate": 100.0,
                    "Total Mole Fraction benzene": 0.5,
                    "Total Mole Fraction toluene": 0.5,
                    "Temperature": 300,
                    "Pressure": 1e5,
                },
                "Cold Side Inlet": {
                    "Total Molar Flowrate": 1.0,
                    "Total Mole Fraction benzene": 0.5,
                    "Total Mole Fraction toluene": 0.5,
                    "Temperature": 300,
                    "Pressure": 101325.0,
                },
                "Cold Side Outlet": {
                    "Total Molar Flowrate": 100.0,
                    "Total Mole Fraction benzene": 0.5,
                    "Total Mole Fraction toluene": 0.5,
                    "Temperature": 300,
                    "Pressure": 1e5,
                },
            }
        )

        pandas.testing.assert_frame_equal(stable, expected, rtol=1e-4, atol=1e-4)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_initialize(self, btx):
        initialization_tester(btx)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solve(self, btx):
        results = solver.solve(btx)

        # Check for optimal solution
        assert check_optimal_termination(results)

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_solution(self, btx):
        assert pytest.approx(5, abs=1e-3) == value(
            btx.fs.unit.hot_side_outlet.flow_mol[0]
        )
        assert pytest.approx(359.4, abs=1e-1) == value(
            btx.fs.unit.hot_side_outlet.temperature[0]
        )
        assert pytest.approx(101325, abs=1e-3) == value(
            btx.fs.unit.hot_side_outlet.pressure[0]
        )

        assert pytest.approx(1, abs=1e-3) == value(
            btx.fs.unit.cold_side_outlet.flow_mol[0]
        )
        assert pytest.approx(596.9, abs=1e-1) == value(
            btx.fs.unit.cold_side_outlet.temperature[0]
        )
        assert pytest.approx(101.325, abs=1e-3) == value(
            btx.fs.unit.cold_side_outlet.pressure[0]
        )

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_conservation(self, btx):
        assert (
            abs(
                value(
                    btx.fs.unit.hot_side_inlet.flow_mol[0]
                    - btx.fs.unit.hot_side_outlet.flow_mol[0]
                )
            )
            <= 1e-6
        )
        assert (
            abs(
                value(
                    btx.fs.unit.cold_side_inlet.flow_mol[0]
                    - btx.fs.unit.cold_side_outlet.flow_mol[0]
                )
            )
            <= 1e-6
        )

        hot_side = value(
            btx.fs.unit.hot_side_outlet.flow_mol[0]
            * (
                btx.fs.unit.hot_side.properties_in[0].enth_mol
                - btx.fs.unit.hot_side.properties_out[0].enth_mol
            )
        )
        cold_side = pyunits.convert_value(
            value(
                btx.fs.unit.cold_side_outlet.flow_mol[0]
                * (
                    btx.fs.unit.cold_side.properties_in[0].enth_mol
                    - btx.fs.unit.cold_side.properties_out[0].enth_mol
                )
            ),
            from_units=pyunits.kJ / pyunits.s,
            to_units=pyunits.J / pyunits.s,
        )
        assert abs(hot_side + cold_side) <= 1e-6

    @pytest.mark.solver
    @pytest.mark.skipif(solver is None, reason="Solver not available")
    @pytest.mark.component
    def test_numerical_issues(self, btx):
        dt = DiagnosticsToolbox(btx)
        # TODO: Complementarity formulation results in near-parallel components
        # when unscaled
        dt.assert_no_numerical_warnings(ignore_parallel_components=True)

    @pytest.mark.component
    def test_initialization_error(self, btx):
        btx.fs.unit.hot_side_outlet.flow_mol[0].fix(20)

        with pytest.raises(InitializationError):
            btx.fs.unit.initialize()


# As we lack other example prop packs with units, take the generic
# BT-PR package and change the base units
configuration2 = {
    # Specifying components
    "components": {
        "benzene": {
            "type": Component,
            "enth_mol_ig_comp": RPP,
            "entr_mol_ig_comp": RPP,
            "pressure_sat_comp": RPP,
            "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            "parameter_data": {
                "mw": (78.1136e-3, pyunits.kg / pyunits.mol),  # [1]
                "pressure_crit": (48.9e5, pyunits.Pa),  # [1]
                "temperature_crit": (562.2, pyunits.K),  # [1]
                "omega": 0.212,  # [1]
                "cp_mol_ig_comp_coeff": {
                    "A": (-3.392e1, pyunits.J / pyunits.mol / pyunits.K),  # [1]
                    "B": (4.739e-1, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (-3.017e-4, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (7.130e-8, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "enth_mol_form_vap_comp_ref": (
                    82.9e3,
                    pyunits.J / pyunits.mol,
                ),  # [3]
                "entr_mol_form_vap_comp_ref": (
                    -269,
                    pyunits.J / pyunits.mol / pyunits.K,
                ),  # [3]
                "pressure_sat_comp_coeff": {
                    "A": (-6.98273, None),  # [1]
                    "B": (1.33213, None),
                    "C": (-2.62863, None),
                    "D": (-3.33399, None),
                },
            },
        },
        "toluene": {
            "type": Component,
            "enth_mol_ig_comp": RPP,
            "entr_mol_ig_comp": RPP,
            "pressure_sat_comp": RPP,
            "phase_equilibrium_form": {("Vap", "Liq"): log_fugacity},
            "parameter_data": {
                "mw": (92.1405e-3, pyunits.kg / pyunits.mol),  # [1]
                "pressure_crit": (41e5, pyunits.Pa),  # [1]
                "temperature_crit": (591.8, pyunits.K),  # [1]
                "omega": 0.263,  # [1]
                "cp_mol_ig_comp_coeff": {
                    "A": (-2.435e1, pyunits.J / pyunits.mol / pyunits.K),  # [1]
                    "B": (5.125e-1, pyunits.J / pyunits.mol / pyunits.K**2),
                    "C": (-2.765e-4, pyunits.J / pyunits.mol / pyunits.K**3),
                    "D": (4.911e-8, pyunits.J / pyunits.mol / pyunits.K**4),
                },
                "enth_mol_form_vap_comp_ref": (
                    50.1e3,
                    pyunits.J / pyunits.mol,
                ),  # [3]
                "entr_mol_form_vap_comp_ref": (
                    -321,
                    pyunits.J / pyunits.mol / pyunits.K,
                ),  # [3]
                "pressure_sat_comp_coeff": {
                    "A": (-7.28607, None),  # [1]
                    "B": (1.38091, None),
                    "C": (-2.83433, None),
                    "D": (-2.79168, None),
                },
            },
        },
    },
    # Specifying phases
    "phases": {
        "Liq": {
            "type": LiquidPhase,
            "equation_of_state": Cubic,
            "equation_of_state_options": {"type": CubicType.PR},
        },
        "Vap": {
            "type": VaporPhase,
            "equation_of_state": Cubic,
            "equation_of_state_options": {"type": CubicType.PR},
        },
    },
    # Set base units of measurement
    "base_units": {
        "time": pyunits.s,
        "length": pyunits.m,
        "mass": pyunits.t,
        "amount": pyunits.mol,
        "temperature": pyunits.degR,
    },
    # Specifying state definition
    "state_definition": FTPx,
    "state_bounds": {
        "flow_mol": (0, 100, 1000, pyunits.mol / pyunits.s),
        "temperature": (273.15, 300, 500, pyunits.K),
        "pressure": (5e4, 1e5, 1e6, pyunits.Pa),
    },
    "pressure_ref": (101325, pyunits.Pa),
    "temperature_ref": (298.15, pyunits.K),
    # Defining phase equilibria
    "phases_in_equilibrium": [("Vap", "Liq")],
    "phase_equilibrium_state": {("Vap", "Liq"): SmoothVLE},
    "bubble_dew_method": LogBubbleDew,
    "parameter_data": {
        "PR_kappa": {
            ("benzene", "benzene"): 0.000,
            ("benzene", "toluene"): 0.000,
            ("toluene", "benzene"): 0.000,
            ("toluene", "toluene"): 0.000,
        }
    },
}


class TestInitializersModular:
    @pytest.mark.integration
    def test_hx0d_initializer(self):
        model = ConcreteModel()
        model.fs = FlowsheetBlock(dynamic=False)

        model.fs.properties = GenericParameterBlock(**configuration)
        model.fs.properties2 = GenericParameterBlock(**configuration2)

        model.fs.unit = HeatExchanger(
            hot_side={"property_package": model.fs.properties},
            cold_side={"property_package": model.fs.properties2},
            flow_pattern=HeatExchangerFlowPattern.cocurrent,
        )

        model.fs.unit.hot_side_inlet.flow_mol[0].fix(5)  # mol/s
        model.fs.unit.hot_side_inlet.temperature[0].fix(365)  # K
        model.fs.unit.hot_side_inlet.pressure[0].fix(101325)  # Pa
        model.fs.unit.hot_side_inlet.mole_frac_comp[0, "benzene"].fix(0.5)
        model.fs.unit.hot_side_inlet.mole_frac_comp[0, "toluene"].fix(0.5)

        model.fs.unit.cold_side_inlet.flow_mol[0].fix(1)  # mol/s
        model.fs.unit.cold_side_inlet.temperature[0].fix(540)  # degR
        model.fs.unit.cold_side_inlet.pressure[0].fix(101.325)  # kPa
        model.fs.unit.cold_side_inlet.mole_frac_comp[0, "benzene"].fix(0.5)
        model.fs.unit.cold_side_inlet.mole_frac_comp[0, "toluene"].fix(0.5)

        model.fs.unit.area.fix(1)
        model.fs.unit.overall_heat_transfer_coefficient.fix(100)

        model.fs.unit.cold_side.scaling_factor_pressure = 1

        # Set small values of epsilon to get sufficiently accurate results
        # Only applies to hot side, as cold side used the original SmoothVLE.
        model.fs.unit.hot_side.properties_in[0].eps_t_Vap_Liq.set_value(1e-4)
        model.fs.unit.hot_side.properties_in[0].eps_z_Vap_Liq.set_value(1e-4)
        model.fs.unit.hot_side.properties_out[0].eps_t_Vap_Liq.set_value(1e-4)
        model.fs.unit.hot_side.properties_out[0].eps_z_Vap_Liq.set_value(1e-4)

        initializer = HX0DInitializer()
        initializer.initialize(model.fs.unit)

        assert initializer.summary[model.fs.unit]["status"] == InitializationStatus.Ok

        assert pytest.approx(5, abs=1e-3) == value(
            model.fs.unit.hot_side_outlet.flow_mol[0]
        )
        assert pytest.approx(359.4, abs=1e-1) == value(
            model.fs.unit.hot_side_outlet.temperature[0]
        )
        assert pytest.approx(101325, abs=1e-3) == value(
            model.fs.unit.hot_side_outlet.pressure[0]
        )

        assert pytest.approx(1, abs=1e-3) == value(
            model.fs.unit.cold_side_outlet.flow_mol[0]
        )
        assert pytest.approx(596.9, abs=1e-1) == value(
            model.fs.unit.cold_side_outlet.temperature[0]
        )
        assert pytest.approx(101.325, abs=1e-3) == value(
            model.fs.unit.cold_side_outlet.pressure[0]
        )

    @pytest.mark.integration
    def test_block_triangularization(
        self,
    ):
        # Trying to get this to work with CubicComplementarityVLE is challenging, and
        # not necessary for this particular test
        new_config = deepcopy(configuration)
        new_config["phase_equilibrium_state"] = {("Vap", "Liq"): SmoothVLE}

        model = ConcreteModel()
        model.fs = FlowsheetBlock(dynamic=False)

        model.fs.properties = GenericParameterBlock(**new_config)
        model.fs.properties2 = GenericParameterBlock(**configuration2)

        model.fs.unit = HeatExchanger(
            hot_side={"property_package": model.fs.properties},
            cold_side={"property_package": model.fs.properties2},
            flow_pattern=HeatExchangerFlowPattern.cocurrent,
        )

        model.fs.unit.hot_side_inlet.flow_mol[0].fix(5)  # mol/s
        model.fs.unit.hot_side_inlet.temperature[0].fix(365)  # K
        model.fs.unit.hot_side_inlet.pressure[0].fix(101325)  # Pa
        model.fs.unit.hot_side_inlet.mole_frac_comp[0, "benzene"].fix(0.5)
        model.fs.unit.hot_side_inlet.mole_frac_comp[0, "toluene"].fix(0.5)

        model.fs.unit.cold_side_inlet.flow_mol[0].fix(1)  # mol/s
        model.fs.unit.cold_side_inlet.temperature[0].fix(540)  # degR
        model.fs.unit.cold_side_inlet.pressure[0].fix(101.325)  # kPa
        model.fs.unit.cold_side_inlet.mole_frac_comp[0, "benzene"].fix(0.5)
        model.fs.unit.cold_side_inlet.mole_frac_comp[0, "toluene"].fix(0.5)

        model.fs.unit.area.fix(1)
        model.fs.unit.overall_heat_transfer_coefficient.fix(100)

        model.fs.unit.cold_side.scaling_factor_pressure = 1

        initializer = BlockTriangularizationInitializer(
            constraint_tolerance=2e-5,
        )
        initializer.initialize(model.fs.unit)

        assert initializer.summary[model.fs.unit]["status"] == InitializationStatus.Ok

        assert pytest.approx(5, abs=1e-3) == value(
            model.fs.unit.hot_side_outlet.flow_mol[0]
        )
        assert pytest.approx(359.4, abs=1e-1) == value(
            model.fs.unit.hot_side_outlet.temperature[0]
        )
        assert pytest.approx(101325, abs=1e-3) == value(
            model.fs.unit.hot_side_outlet.pressure[0]
        )

        assert pytest.approx(1, abs=1e-3) == value(
            model.fs.unit.cold_side_outlet.flow_mol[0]
        )
        assert pytest.approx(596.9, abs=1e-1) == value(
            model.fs.unit.cold_side_outlet.temperature[0]
        )
        assert pytest.approx(101.325, abs=1e-3) == value(
            model.fs.unit.cold_side_outlet.pressure[0]
        )


class TestInitializersIAPWS:
    @pytest.fixture
    def model(self):
        m = ConcreteModel()
        m.fs = FlowsheetBlock(dynamic=False)

        m.fs.properties = iapws95.Iapws95ParameterBlock()

        m.fs.unit = HeatExchanger(
            hot_side={"property_package": m.fs.properties},
            cold_side={"property_package": m.fs.properties},
            flow_pattern=HeatExchangerFlowPattern.countercurrent,
        )

        m.fs.unit.hot_side_inlet.flow_mol[0].set_value(100)
        m.fs.unit.hot_side_inlet.enth_mol[0].set_value(6000)
        m.fs.unit.hot_side_inlet.pressure[0].set_value(101325)

        m.fs.unit.cold_side_inlet.flow_mol[0].set_value(100)
        m.fs.unit.cold_side_inlet.enth_mol[0].set_value(5000)
        m.fs.unit.cold_side_inlet.pressure[0].set_value(101325)

        m.fs.unit.area.fix(1000)
        m.fs.unit.overall_heat_transfer_coefficient.fix(100)

        return m

    @pytest.mark.component
    def test_hx0d_initializer(self, model):
        initializer = HX0DInitializer()
        initializer.initialize(model.fs.unit)

        assert initializer.summary[model.fs.unit]["status"] == InitializationStatus.Ok

        assert pytest.approx(100, abs=1e-5) == value(
            model.fs.unit.hot_side_outlet.flow_mol[0]
        )
        assert pytest.approx(100, abs=1e-5) == value(
            model.fs.unit.cold_side_outlet.flow_mol[0]
        )

        assert pytest.approx(5070.219, abs=1e0) == value(
            model.fs.unit.hot_side_outlet.enth_mol[0]
        )
        assert pytest.approx(5929.781, abs=1e0) == value(
            model.fs.unit.cold_side_outlet.enth_mol[0]
        )

        assert pytest.approx(101325, abs=1e2) == value(
            model.fs.unit.hot_side_outlet.pressure[0]
        )
        assert pytest.approx(101325, abs=1e2) == value(
            model.fs.unit.cold_side_outlet.pressure[0]
        )

        assert not model.fs.unit.hot_side_inlet.flow_mol[0].fixed
        assert not model.fs.unit.hot_side_inlet.enth_mol[0].fixed
        assert not model.fs.unit.hot_side_inlet.pressure[0].fixed

        assert not model.fs.unit.cold_side_inlet.flow_mol[0].fixed
        assert not model.fs.unit.cold_side_inlet.enth_mol[0].fixed
        assert not model.fs.unit.cold_side_inlet.pressure[0].fixed

    @pytest.mark.component
    def test_block_triangularization(self, model):
        initializer = BlockTriangularizationInitializer(constraint_tolerance=2e-5)
        initializer.initialize(model.fs.unit)

        assert initializer.summary[model.fs.unit]["status"] == InitializationStatus.Ok

        assert pytest.approx(100, abs=1e-5) == value(
            model.fs.unit.hot_side_outlet.flow_mol[0]
        )
        assert pytest.approx(100, abs=1e-5) == value(
            model.fs.unit.cold_side_outlet.flow_mol[0]
        )

        assert pytest.approx(5070.219, abs=1e0) == value(
            model.fs.unit.hot_side_outlet.enth_mol[0]
        )
        assert pytest.approx(5929.781, abs=1e0) == value(
            model.fs.unit.cold_side_outlet.enth_mol[0]
        )

        assert pytest.approx(101325, abs=1e2) == value(
            model.fs.unit.hot_side_outlet.pressure[0]
        )
        assert pytest.approx(101325, abs=1e2) == value(
            model.fs.unit.cold_side_outlet.pressure[0]
        )

        assert not model.fs.unit.hot_side_inlet.flow_mol[0].fixed
        assert not model.fs.unit.hot_side_inlet.enth_mol[0].fixed
        assert not model.fs.unit.hot_side_inlet.pressure[0].fixed

        assert not model.fs.unit.cold_side_inlet.flow_mol[0].fixed
        assert not model.fs.unit.cold_side_inlet.enth_mol[0].fixed
        assert not model.fs.unit.cold_side_inlet.pressure[0].fixed



#################################################################################
# The Institute for the Design of Advanced Energy Systems Integrated Platform
# Framework (IDAES IP) was produced under the DOE Institute for the
# Design of Advanced Energy Systems (IDAES).
#
# Copyright (c) 2018-2024 by the software owners: The Regents of the
# University of California, through Lawrence Berkeley National Laboratory,
# National Technology & Engineering Solutions of Sandia, LLC, Carnegie Mellon
# University, West Virginia University Research Corporation, et al.
# All rights reserved.  Please see the files COPYRIGHT.md and LICENSE.md
# for full copyright and license information.
#################################################################################
"""
Ideal Liquid Phase properties for aqueous MEA solvent with CO2.

The following apparent species are used to represent the mixture, along with
the method used to calculate their vapor pressure:

    Carbon Dioxide (CO2) - Henry's Law
    Monoethanolamine (MEA) - non-volatile,
    Water (H2O) - Raoult's Law

Additionally, the following true ionic species are required for calculating
transport properties:

    MEA+, MEACOO-, HCO3-

Assumptions:
    * No heats of formation
    * CO2 contribution to enthalpy is due to heat of absorption

References:
    [1] Hilliard thesis (1998)
    [2] Morgan et.al (2015)
    [3] NIST Webbook, https://webbook.nist.gov/
"""
# TODO: Missing docstrings
# pylint: disable=missing-class-docstring
# pylint: disable=missing-function-docstring

# Import Pyomo units
from pyomo.environ import exp, log, units as pyunits, Var, Expression

# Import IDAES cores
from idaes.core import AqueousPhase, Solvent, Solute, Anion, Cation

from idaes.models.properties.modular_properties.state_definitions import FTPx
from idaes.models.properties.modular_properties.base.generic_property import StateIndex
from idaes.models.properties.modular_properties.eos.ideal import Ideal

from idaes.models.properties.modular_properties.reactions.equilibrium_forms import (
    log_power_law_equil,
)
from idaes.models.properties.modular_properties.base.utility import ConcentrationForm
from idaes.models.properties.modular_properties.phase_equil.henry import HenryType

from idaes.core.util.misc import set_param_from_config
import idaes.logger as idaeslog


# Set up logger
_log = idaeslog.getLogger(__name__)


# -----------------------------------------------------------------------------
# Pure Component Property methods for aqueous MEA
class CpMolCO2:
    # No contribution to enthalpy from CO2, but need to add a contribution
    # to cp to avoid artificial decrease in as CO2 is absorbed.
    # Assume cp contribution of CO2 is equal to weighted average of cp for
    # solvents
    @staticmethod
    def return_expression(b, cobj, T):
        return (
            (
                b.mole_frac_phase_comp["Liq", "H2O"] * b.cp_mol_phase_comp["Liq", "H2O"]
                + b.mole_frac_phase_comp["Liq", "MEA"]
                * b.cp_mol_phase_comp["Liq", "MEA"]
            )
            / b.mole_frac_phase_comp["Liq", "CO2"]
            * (1 / (1 - b.mass_frac_phase_comp["Liq", "CO2"]) - 1)
        )


class CpMolSolvent:
    # Method to calculate cp for solvents [1]
    @staticmethod
    def build_parameters(cobj):
        cobj.cp_mass_liq_comp_coeff_1 = Var(
            doc="Parameter 1 for liquid phase mass specific heat capacity",
            units=pyunits.kJ * pyunits.kg**-1 * pyunits.K**-1,
        )
        set_param_from_config(cobj, param="cp_mass_liq_comp_coeff", index="1")

        cobj.cp_mass_liq_comp_coeff_2 = Var(
            doc="Parameter 2 for liquid phase mass specific heat capacity",
            units=pyunits.kJ * pyunits.kg**-1 * pyunits.K**-2,
        )
        set_param_from_config(cobj, param="cp_mass_liq_comp_coeff", index="2")

        cobj.cp_mass_liq_comp_coeff_3 = Var(
            doc="Parameter 3 for liquid phase mass specific heat capacity",
            units=pyunits.kJ * pyunits.kg**-1 * pyunits.K**-3,
        )
        set_param_from_config(cobj, param="cp_mass_liq_comp_coeff", index="3")

        cobj.cp_mass_liq_comp_coeff_4 = Var(
            doc="Parameter 4 for liquid phase mass specific heat capacity",
            units=pyunits.kJ * pyunits.kg**-1 * pyunits.K**-4,
        )
        set_param_from_config(cobj, param="cp_mass_liq_comp_coeff", index="4")

        cobj.cp_mass_liq_comp_coeff_5 = Var(
            doc="Parameter 5 for liquid phase mass specific heat capacity",
            units=pyunits.kJ * pyunits.kg**-1 * pyunits.K**-5,
        )
        set_param_from_config(cobj, param="cp_mass_liq_comp_coeff", index="5")

    @staticmethod
    def return_expression(b, cobj, T):
        # Specific heat capacity
        # Need to convert K to C - fake units for now.
        T = pyunits.convert(T, to_units=pyunits.K) - 273.15 * pyunits.K
        cp = cobj.mw * (
            cobj.cp_mass_liq_comp_coeff_5 * T**4
            + cobj.cp_mass_liq_comp_coeff_4 * T**3
            + cobj.cp_mass_liq_comp_coeff_3 * T**2
            + cobj.cp_mass_liq_comp_coeff_2 * T
            + cobj.cp_mass_liq_comp_coeff_1
        )

        units = b.params.get_metadata().derived_units
        return pyunits.convert(cp, units.HEAT_CAPACITY_MOLE)


class EnthMolCO2:
    # Only contribution to enthalpy is heat of absorption
    @staticmethod
    def build_parameters(cobj):
        cobj.dh_abs_co2 = Var(
            doc="Heat of absorption of CO2 @ Tref", units=pyunits.J / pyunits.mol
        )
        set_param_from_config(cobj, param="dh_abs_co2")

    @staticmethod
    def return_expression(b, cobj, T):
        return cobj.dh_abs_co2


class EnthMolSolvent:
    # Method to calculate specific enthalpy of solvents
    @staticmethod
    def build_parameters(cobj):
        if not hasattr(cobj, "cp_mass_liq_comp_coeff_1"):
            CpMolSolvent.build_parameters(cobj)

        cobj.dh_vap = Var(
            doc="Heat of vaporization of component @ Tref",
            units=pyunits.J / pyunits.mol,
        )
        set_param_from_config(cobj, param="dh_vap")

    @staticmethod
    def return_expression(b, cobj, T):
        # Specific enthalpy
        T = pyunits.convert(T, to_units=pyunits.K) - 273.15 * pyunits.K
        Tr = (
            pyunits.convert(b.params.temperature_ref, to_units=pyunits.K)
            - 273.15 * pyunits.K
        )

        units = b.params.get_metadata().derived_units

        h = (
            pyunits.convert(
                cobj.mw
                * (
                    (cobj.cp_mass_liq_comp_coeff_5 / 5) * (T**5 - Tr**5)
                    + (cobj.cp_mass_liq_comp_coeff_4 / 4) * (T**4 - Tr**4)
                    + (cobj.cp_mass_liq_comp_coeff_3 / 3) * (T**3 - Tr**3)
                    + (cobj.cp_mass_liq_comp_coeff_2 / 2) * (T**2 - Tr**2)
                    + cobj.cp_mass_liq_comp_coeff_1 * (T - Tr)
                ),
                units.ENERGY_MOLE,
            )
            - cobj.dh_vap
        )

        return h


class N2OAnalogy:
    # Henry's constant N2O Analogy Jiru et.al (2012)
    # Units of original expression are Pa*m^3/mol
    # TODO: Handle units of H
    @staticmethod
    def build_parameters(cobj, phase, h_type):
        cobj.lwm_coeff_1 = Var(
            doc="N2O Analogy Henry's constant coefficient 1",
            units=pyunits.dimensionless,
        )
        set_param_from_config(cobj, param="lwm_coeff", index="1")

        cobj.lwm_coeff_2 = Var(
            doc="N2O Analogy Henry's constant coefficient 2",
            units=pyunits.dimensionless,
        )
        set_param_from_config(cobj, param="lwm_coeff", index="2")

        cobj.lwm_coeff_3 = Var(
            doc="N2O Analogy Henry's constant coefficient 3",
            units=pyunits.dimensionless,
        )
        set_param_from_config(cobj, param="lwm_coeff", index="3")

        cobj.lwm_coeff_4 = Var(
            doc="N2O Analogy Henry's constant coefficient 4",
            units=pyunits.dimensionless,
        )
        set_param_from_config(cobj, param="lwm_coeff", index="4")

    @staticmethod
    def return_expression(b, p, j, T):
        cobj = b.params.get_component(j)
        t = T - 273.15 * pyunits.K

        # Calculations require mass fraction of MEA and H2O on a CO2 free basis
        m_MEA = b.mole_frac_comp["MEA"] * b.params.MEA.mw
        m_H2O = b.mole_frac_comp["H2O"] * b.params.H2O.mw
        wt_MEA = m_MEA / (m_MEA + m_H2O)
        wt_H2O = m_H2O / (m_MEA + m_H2O)
        H_N2O_MEA = 2.448e5 * exp(-1348 * pyunits.K / T)
        H_CO2_H2O = 3.52e6 * exp(-2113 * pyunits.K / T)
        H_N2O_H2O = 8.449e6 * exp(-2283 * pyunits.K / T)
        H_CO2_MEA = H_N2O_MEA * (H_CO2_H2O / H_N2O_H2O)
        lwm = (
            cobj.lwm_coeff_1
            + cobj.lwm_coeff_2 * pyunits.K**-1 * t
            + cobj.lwm_coeff_3 * pyunits.K**-2 * t**2
            + cobj.lwm_coeff_4 * wt_H2O
        )

        return (
            (
                exp(
                    wt_MEA * log(H_CO2_MEA)
                    + wt_H2O * log(H_CO2_H2O)
                    + wt_MEA * wt_H2O * lwm
                )
            )
            * pyunits.Pa
            * pyunits.m**3
            * pyunits.mol**-1
        )


class PressureSatSolvent:
    # Method for calculating saturation pressure ofsolvents
    @staticmethod
    def build_parameters(cobj):
        cobj.pressure_sat_comp_coeff_1 = Var(
            doc="Coefficient 1 for calculating Psat", units=pyunits.dimensionless
        )
        set_param_from_config(cobj, param="pressure_sat_comp_coeff", index="1")

        cobj.pressure_sat_comp_coeff_2 = Var(
            doc="Coefficient 2 for calculating Psat", units=pyunits.K
        )
        set_param_from_config(cobj, param="pressure_sat_comp_coeff", index="2")

        cobj.pressure_sat_comp_coeff_3 = Var(
            doc="Coefficient 3 for calculating Psat", units=pyunits.dimensionless
        )
        set_param_from_config(cobj, param="pressure_sat_comp_coeff", index="3")

        cobj.pressure_sat_comp_coeff_4 = Var(
            doc="Coefficient 4 for calculating Psat", units=pyunits.K**-2
        )
        set_param_from_config(cobj, param="pressure_sat_comp_coeff", index="4")

    @staticmethod
    def return_expression(b, cobj, T, dT=False):
        if dT:
            # PYLINT-TODO
            # pylint: disable-next=broad-exception-raised
            raise Exception("No dT method for pressure sat")

        return (
            exp(
                cobj.pressure_sat_comp_coeff_1
                + cobj.pressure_sat_comp_coeff_2 / T
                + cobj.pressure_sat_comp_coeff_3 * log(T / pyunits.K)
                + cobj.pressure_sat_comp_coeff_4 * T**2
            )
        ) * pyunits.Pa


class VolMolSolvent:
    # Weiland Method for calculating molar volume of pure solvents [2]

    @staticmethod
    def build_parameters(cobj):
        cobj.dens_mol_liq_comp_coeff_1 = Var(
            doc="Parameter 1 for liquid phase molar density",
            units=pyunits.g / pyunits.mL / pyunits.K**2,
        )
        set_param_from_config(cobj, param="dens_mol_liq_comp_coeff", index="1")

        cobj.dens_mol_liq_comp_coeff_2 = Var(
            doc="Parameter 2 for liquid phase molar density",
            units=pyunits.g / pyunits.mL / pyunits.K,
        )
        set_param_from_config(cobj, param="dens_mol_liq_comp_coeff", index="2")

        cobj.dens_mol_liq_comp_coeff_3 = Var(
            doc="Parameter 3 for liquid phase molar density",
            units=pyunits.g / pyunits.mL,
        )
        set_param_from_config(cobj, param="dens_mol_liq_comp_coeff", index="3")

    @staticmethod
    def return_expression(b, cobj, T):
        T = pyunits.convert(T, to_units=pyunits.K)

        rho = (
            cobj.dens_mol_liq_comp_coeff_1 * T**2
            + cobj.dens_mol_liq_comp_coeff_2 * T
            + cobj.dens_mol_liq_comp_coeff_3
        )
        vol_mol = cobj.mw / rho

        units = b.params.get_metadata().derived_units

        return pyunits.convert(vol_mol, units.VOLUME / units.AMOUNT)


class VolMolCO2:
    # Weiland Method for calculating molar volume of dissolved CO2 [2]

    @staticmethod
    def build_parameters(cobj):
        cobj.vol_mol_liq_comp_coeff_a = Var(
            doc="Parameter a for liquid phase molar volume",
            units=pyunits.mL / pyunits.mol,
        )
        set_param_from_config(cobj, param="vol_mol_liq_comp_coeff", index="a")

        cobj.vol_mol_liq_comp_coeff_b = Var(
            doc="Parameter b for liquid phase molar volume",
            units=pyunits.mL / pyunits.mol,
        )
        set_param_from_config(cobj, param="vol_mol_liq_comp_coeff", index="b")

        cobj.vol_mol_liq_comp_coeff_c = Var(
            doc="Parameter c for liquid phase molar volume",
            units=pyunits.mL / pyunits.mol,
        )
        set_param_from_config(cobj, param="vol_mol_liq_comp_coeff", index="c")

        cobj.vol_mol_liq_comp_coeff_d = Var(
            doc="Parameter d for liquid phase molar volume",
            units=pyunits.mL / pyunits.mol,
        )
        set_param_from_config(cobj, param="vol_mol_liq_comp_coeff", index="d")

        cobj.vol_mol_liq_comp_coeff_e = Var(
            doc="Parameter e for liquid phase molar volume",
            units=pyunits.mL / pyunits.mol,
        )
        set_param_from_config(cobj, param="vol_mol_liq_comp_coeff", index="e")

    @staticmethod
    def return_expression(b, cobj, T):
        x = b.mole_frac_comp

        vol_mol = (
            cobj.vol_mol_liq_comp_coeff_a
            + (cobj.vol_mol_liq_comp_coeff_b + cobj.vol_mol_liq_comp_coeff_c * x["MEA"])
            * x["MEA"]
            * x["H2O"]
            / x["CO2"]
            + (cobj.vol_mol_liq_comp_coeff_d + cobj.vol_mol_liq_comp_coeff_e * x["MEA"])
            * x["MEA"]
        )

        units = b.params.get_metadata().derived_units

        return pyunits.convert(vol_mol, units.VOLUME / units.AMOUNT)


# -----------------------------------------------------------------------------
# Transport property models
class DiffusCO2:
    @staticmethod
    def build_parameters(cobj, phase):
        cobj.diffus_phase_comp_coeff_1 = Var(
            doc="Parameter 1 for liquid phase diffusivity model",
            units=pyunits.m**2 / pyunits.s,
        )
        set_param_from_config(cobj, param="diffus_phase_comp_coeff", index="1")
        cobj.diffus_phase_comp_coeff_2 = Var(
            doc="Parameter 2 for liquid phase diffusivity model",
            units=pyunits.m**5 / pyunits.kmol / pyunits.s,
        )
        set_param_from_config(cobj, param="diffus_phase_comp_coeff", index="2")
        cobj.diffus_phase_comp_coeff_3 = Var(
            doc="Parameter 3 for liquid phase diffusivity model",
            units=pyunits.m**8 / pyunits.kmol**2 / pyunits.s,
        )
        set_param_from_config(cobj, param="diffus_phase_comp_coeff", index="3")
        cobj.diffus_phase_comp_coeff_4 = Var(
            doc="Parameter 4 for liquid phase diffusivity model", units=pyunits.K
        )
        set_param_from_config(cobj, param="diffus_phase_comp_coeff", index="4")
        cobj.diffus_phase_comp_coeff_5 = Var(
            doc="Parameter 5 for liquid phase diffusivity model",
            units=(pyunits.m**3) * pyunits.K / pyunits.kmol,
        )
        set_param_from_config(cobj, param="diffus_phase_comp_coeff", index="5")

    @staticmethod
    def return_expression(blk, p, j, T):
        cobj = blk.params.get_component(j)
        C_MEA = blk.conc_mol_comp["MEA"] * 1e-3 * pyunits.kmol / pyunits.mol
        return (
            cobj.diffus_phase_comp_coeff_1
            + cobj.diffus_phase_comp_coeff_2 * C_MEA
            + cobj.diffus_phase_comp_coeff_3 * C_MEA**2
        ) * exp(
            (cobj.diffus_phase_comp_coeff_4 + (cobj.diffus_phase_comp_coeff_5 * C_MEA))
            / T
        )


class DiffusMEA:
    @staticmethod
    def build_parameters(cobj, phase):
        cobj.diffus_phase_comp_coeff_1 = Var(
            doc="Parameter 1 for liquid phase diffusivity model",
            units=pyunits.dimensionless,
        )
        set_param_from_config(cobj, param="diffus_phase_comp_coeff", index="1")
        cobj.diffus_phase_comp_coeff_2 = Var(
            doc="Parameter 2 for liquid phase diffusivity model", units=pyunits.K
        )
        set_param_from_config(cobj, param="diffus_phase_comp_coeff", index="2")
        cobj.diffus_phase_comp_coeff_3 = Var(
            doc="Parameter 3 for liquid phase diffusivity model",
            units=pyunits.m**3 / pyunits.mol,
        )
        set_param_from_config(cobj, param="diffus_phase_comp_coeff", index="3")

    @staticmethod
    def return_expression(blk, p, j, T):
        cobj = blk.params.get_component(j)
        C_MEA = blk.conc_mol_comp["MEA"]
        return (
            exp(
                cobj.diffus_phase_comp_coeff_1
                + cobj.diffus_phase_comp_coeff_2 / T
                + cobj.diffus_phase_comp_coeff_3 * C_MEA
            )
            * pyunits.m**2
            / pyunits.s
        )


class DiffusIons:
    @staticmethod
    def build_parameters(cobj, phase):
        cobj.diffus_phase_comp_coeff_1 = Var(
            doc="Parameter 1 for liquid phase diffusivity model",
            units=pyunits.dimensionless,
        )
        set_param_from_config(cobj, param="diffus_phase_comp_coeff", index="1")
        cobj.diffus_phase_comp_coeff_2 = Var(
            doc="Parameter 2 for liquid phase diffusivity model", units=pyunits.K
        )
        set_param_from_config(cobj, param="diffus_phase_comp_coeff", index="2")
        cobj.diffus_phase_comp_coeff_3 = Var(
            doc="Parameter 3 for liquid phase diffusivity model",
            units=pyunits.dimensionless,
        )
        set_param_from_config(cobj, param="diffus_phase_comp_coeff", index="3")

    @staticmethod
    def return_expression(blk, p, j, T):
        cobj = blk.params.get_component(j)
        return (
            exp(
                cobj.diffus_phase_comp_coeff_1
                + cobj.diffus_phase_comp_coeff_2 / T
                + cobj.diffus_phase_comp_coeff_3
                * log(blk.visc_d_phase[p] / pyunits.get_units(blk.visc_d_phase[p]))
            )
            * pyunits.m**2
            / pyunits.s
        )


class DiffusNone:
    # placeholder method for components where diffusivity is not required
    @staticmethod
    def build_parameters(cobj, phase):
        pass

    @staticmethod
    def return_expression(blk, p, j, T):
        return Expression.Skip


class Viscosity:
    @staticmethod
    def build_parameters(pobj):
        pobj.visc_d_coeff_a = Var(
            doc="Parameter a for liquid phase viscosity model", units=pyunits.K**-1
        )
        set_param_from_config(pobj, param="visc_d_coeff", index="a")

        pobj.visc_d_coeff_b = Var(
            doc="Parameter b for liquid phase viscosity model", units=pyunits.K**-1
        )
        set_param_from_config(pobj, param="visc_d_coeff", index="b")

        pobj.visc_d_coeff_c = Var(
            doc="Parameter c for liquid phase viscosity model",
            units=pyunits.dimensionless,
        )
        set_param_from_config(pobj, param="visc_d_coeff", index="c")

        pobj.visc_d_coeff_d = Var(
            doc="Parameter d for liquid phase viscosity model",
            units=pyunits.dimensionless,
        )
        set_param_from_config(pobj, param="visc_d_coeff", index="d")

        pobj.visc_d_coeff_e = Var(
            doc="Parameter e for liquid phase viscosity model",
            units=pyunits.dimensionless,
        )
        set_param_from_config(pobj, param="visc_d_coeff", index="e")

        pobj.visc_d_coeff_f = Var(
            doc="Parameter f for liquid phase viscosity model", units=pyunits.K**-1
        )
        set_param_from_config(pobj, param="visc_d_coeff", index="f")

        pobj.visc_d_coeff_g = Var(
            doc="Parameter g for liquid phase viscosity model",
            units=pyunits.dimensionless,
        )
        set_param_from_config(pobj, param="visc_d_coeff", index="g")

    @staticmethod
    def return_expression(blk, phase):
        pobj = blk.params.get_phase(phase)

        r = (
            blk.mass_frac_phase_comp_apparent["Liq", "MEA"]
            / (
                blk.mass_frac_phase_comp_apparent["Liq", "MEA"]
                + blk.mass_frac_phase_comp_apparent["Liq", "H2O"]
            )
        ) * 100
        T = blk.temperature
        alpha = (
            blk.mole_frac_phase_comp_apparent["Liq", "CO2"]
            / blk.mole_frac_phase_comp_apparent["Liq", "MEA"]
        )
        mu_H2O = (
            1.002e-3
            * pyunits.Pa
            * pyunits.s
            * 10
            ** (
                (
                    1.3272
                    * (
                        293.15 * pyunits.K
                        - T
                        - 0.001053 * pyunits.K**-1 * (T - 293.15 * pyunits.K) ** 2
                    )
                )
                / (T - 168.15 * pyunits.K)
            )
        )
        a = pobj.visc_d_coeff_a
        b = pobj.visc_d_coeff_b
        c = pobj.visc_d_coeff_c
        d = pobj.visc_d_coeff_d
        e = pobj.visc_d_coeff_e
        f = pobj.visc_d_coeff_f
        g = pobj.visc_d_coeff_g

        # Model appears to be entirely empirical, and units are not obvious
        # Assume each part of expression is unitless and use units to match
        return mu_H2O * exp(
            r
            * (T * (a * r + b) + c * r + d)
            * (alpha * (e * r + f * T + g) + 1)
            / T**2
            * pyunits.K**2
        )


class ThermalCond:
    @staticmethod
    def build_parameters(pobj):
        pass

    @staticmethod
    def return_expression(blk, phase):
        Tb_MEA = 443 * pyunits.K
        Tc_MEA = 614.2 * pyunits.K
        T = blk.temperature
        Tr = T / Tc_MEA
        Tbr_MEA = Tb_MEA / Tc_MEA
        K_MEA = (
            1.1053152
            / (61.08**0.5)
            * (3 + 20 * (1 - Tr) ** (2 / 3))
            / (3 + 20 * (1 - Tbr_MEA) ** (2 / 3))
        )
        K_H2O = 0.6065 * (
            -1.48445
            + 4.12292 * T / (298.15 * pyunits.K)
            - 1.63866 * (T / (298.15 * pyunits.K)) ** 2
        )

        x = blk.mole_frac_phase_comp_apparent
        return (
            ((x["Liq", "H2O"] * K_H2O**-2 + x["Liq", "MEA"] * K_MEA**-2) ** -1) ** 0.5
            * pyunits.W
            / pyunits.m
            / pyunits.K
        )


class SurfTens:
    @staticmethod
    def build_parameters(pobj):
        pobj.surf_tens_H2O_coeff_1 = Var(
            doc="Parameter 1 for surface tension model of pure water",
            units=pyunits.N / pyunits.m,
        )
        set_param_from_config(pobj, param="surf_tens_H2O_coeff", index="1")
        pobj.surf_tens_H2O_coeff_2 = Var(
            doc="Parameter 2 for surface tension model of pure water",
            units=pyunits.dimensionless,
        )
        set_param_from_config(pobj, param="surf_tens_H2O_coeff", index="2")
        pobj.surf_tens_H2O_coeff_3 = Var(
            doc="Parameter 3 for surface tension model of pure water",
            units=pyunits.dimensionless,
        )
        set_param_from_config(pobj, param="surf_tens_H2O_coeff", index="3")
        pobj.surf_tens_H2O_coeff_4 = Var(
            doc="Parameter 4 for surface tension model of pure water",
            units=pyunits.dimensionless,
        )
        set_param_from_config(pobj, param="surf_tens_H2O_coeff", index="4")

        pobj.surf_tens_MEA_coeff_1 = Var(
            doc="Parameter 1 for surface tension model of pure MEA",
            units=pyunits.N / pyunits.m,
        )
        set_param_from_config(pobj, param="surf_tens_MEA_coeff", index="1")
        pobj.surf_tens_MEA_coeff_2 = Var(
            doc="Parameter 2 for surface tension model of pure MEA",
            units=pyunits.dimensionless,
        )
        set_param_from_config(pobj, param="surf_tens_MEA_coeff", index="2")
        pobj.surf_tens_MEA_coeff_3 = Var(
            doc="Parameter 3 for surface tension model of pure MEA",
            units=pyunits.dimensionless,
        )
        set_param_from_config(pobj, param="surf_tens_MEA_coeff", index="3")
        pobj.surf_tens_MEA_coeff_4 = Var(
            doc="Parameter 4 for surface tension model of pure MEA",
            units=pyunits.dimensionless,
        )
        set_param_from_config(pobj, param="surf_tens_MEA_coeff", index="4")

        pobj.surf_tens_CO2_coeff_1 = Var(
            doc="Parameter 1 for surface tension model of pure CO2",
            units=pyunits.N / pyunits.m,
        )
        set_param_from_config(pobj, param="surf_tens_CO2_coeff", index="1")
        pobj.surf_tens_CO2_coeff_2 = Var(
            doc="Parameter 2 for surface tension model of pure CO2",
            units=pyunits.N / pyunits.m,
        )
        set_param_from_config(pobj, param="surf_tens_CO2_coeff", index="2")
        pobj.surf_tens_CO2_coeff_3 = Var(
            doc="Parameter 3 for surface tension model of pure CO2",
            units=pyunits.N / pyunits.m,
        )
        set_param_from_config(pobj, param="surf_tens_CO2_coeff", index="3")
        pobj.surf_tens_CO2_coeff_4 = Var(
            doc="Parameter 4 for surface tension model of pure CO2",
            units=pyunits.N / pyunits.m / pyunits.K,
        )
        set_param_from_config(pobj, param="surf_tens_CO2_coeff", index="4")
        pobj.surf_tens_CO2_coeff_5 = Var(
            doc="Parameter 5 for surface tension model of pure CO2",
            units=pyunits.N / pyunits.m / pyunits.K,
        )
        set_param_from_config(pobj, param="surf_tens_CO2_coeff", index="5")
        pobj.surf_tens_CO2_coeff_6 = Var(
            doc="Parameter 6 for surface tension model of pure CO2",
            units=pyunits.N / pyunits.m / pyunits.K,
        )
        set_param_from_config(pobj, param="surf_tens_CO2_coeff", index="6")

        pobj.surf_tens_F_coeff_a = Var(
            doc="Parameter Fa for liquid phase surface tension model",
            units=pyunits.dimensionless,
        )
        set_param_from_config(pobj, param="surf_tens_F_coeff", index="a")
        pobj.surf_tens_F_coeff_b = Var(
            doc="Parameter Fb for liquid phase surface tension model",
            units=pyunits.dimensionless,
        )
        set_param_from_config(pobj, param="surf_tens_F_coeff", index="b")
        pobj.surf_tens_F_coeff_c = Var(
            doc="Parameter Fc for liquid phase surface tension model",
            units=pyunits.dimensionless,
        )
        set_param_from_config(pobj, param="surf_tens_F_coeff", index="c")
        pobj.surf_tens_F_coeff_d = Var(
            doc="Parameter Fd for liquid phase surface tension model",
            units=pyunits.dimensionless,
        )
        set_param_from_config(pobj, param="surf_tens_F_coeff", index="d")
        pobj.surf_tens_F_coeff_e = Var(
            doc="Parameter Fe for liquid phase surface tension model",
            units=pyunits.dimensionless,
        )
        set_param_from_config(pobj, param="surf_tens_F_coeff", index="e")
        pobj.surf_tens_F_coeff_f = Var(
            doc="Parameter Ff for liquid phase surface tension model",
            units=pyunits.dimensionless,
        )
        set_param_from_config(pobj, param="surf_tens_F_coeff", index="f")
        pobj.surf_tens_F_coeff_g = Var(
            doc="Parameter Fg for liquid phase surface tension model",
            units=pyunits.dimensionless,
        )
        set_param_from_config(pobj, param="surf_tens_F_coeff", index="g")
        pobj.surf_tens_F_coeff_h = Var(
            doc="Parameter Fh for liquid phase surface tension model",
            units=pyunits.dimensionless,
        )
        set_param_from_config(pobj, param="surf_tens_F_coeff", index="h")
        pobj.surf_tens_F_coeff_i = Var(
            doc="Parameter Fi for liquid phase surface tension model",
            units=pyunits.dimensionless,
        )
        set_param_from_config(pobj, param="surf_tens_F_coeff", index="i")
        pobj.surf_tens_F_coeff_j = Var(
            doc="Parameter Fj for liquid phase surface tension model",
            units=pyunits.dimensionless,
        )
        set_param_from_config(pobj, param="surf_tens_F_coeff", index="j")

    @staticmethod
    def return_expression(blk, phase):
        pobj = blk.params.get_phase(phase)

        T = blk.temperature
        r = blk.mass_frac_phase_comp["Liq", "MEA"] / (
            blk.mass_frac_phase_comp["Liq", "MEA"]
            + blk.mass_frac_phase_comp["Liq", "H2O"]
        )

        alpha = (
            blk.mole_frac_phase_comp_apparent["Liq", "CO2"]
            / blk.mole_frac_phase_comp_apparent["Liq", "MEA"]
        )

        Tc_h2o = blk.params.H2O.temperature_crit
        Tc_mea = blk.params.MEA.temperature_crit

        sigma_h2o = pobj.surf_tens_H2O_coeff_1 * (1 - T / Tc_h2o) ** (
            pobj.surf_tens_H2O_coeff_2
            + pobj.surf_tens_H2O_coeff_3 * T / Tc_h2o
            + pobj.surf_tens_H2O_coeff_4 * (T / Tc_h2o) ** 2
        )
        sigma_mea = pobj.surf_tens_MEA_coeff_1 * (1 - T / Tc_mea) ** (
            pobj.surf_tens_MEA_coeff_2
            + pobj.surf_tens_MEA_coeff_3 * T / Tc_mea
            + pobj.surf_tens_MEA_coeff_4 * (T / Tc_mea) ** 2
        )
        sigma_co2 = (
            pobj.surf_tens_CO2_coeff_1 * r**2
            + pobj.surf_tens_CO2_coeff_2 * r
            + pobj.surf_tens_CO2_coeff_3
            + T
            * (
                pobj.surf_tens_CO2_coeff_4 * r**2
                + pobj.surf_tens_CO2_coeff_5 * r
                + pobj.surf_tens_CO2_coeff_6
            )
        )

        Fa = pobj.surf_tens_F_coeff_a
        Fb = pobj.surf_tens_F_coeff_b
        Fc = pobj.surf_tens_F_coeff_c
        Fd = pobj.surf_tens_F_coeff_d
        Fe = pobj.surf_tens_F_coeff_e
        Ff = pobj.surf_tens_F_coeff_f
        Fg = pobj.surf_tens_F_coeff_g
        Fh = pobj.surf_tens_F_coeff_h
        Fi = pobj.surf_tens_F_coeff_i
        Fj = pobj.surf_tens_F_coeff_j

        return (
            sigma_h2o
            + (sigma_co2 - sigma_h2o)
            * blk.mole_frac_comp["CO2"]
            * (Fa + Fb * alpha + Fc * alpha**2 + Fd * r + Fe * r**2)
            + (sigma_mea - sigma_h2o)
            * (Ff + Fg * alpha + Fh * alpha**2 + Fi * r + Fj * r**2)
            * blk.mole_frac_comp["MEA"]
        )


# -----------------------------------------------------------------------------
# Equilibrium constant model
class k_eq:
    @staticmethod
    def build_parameters(rblock, config):
        rblock.k_eq_coeff_1 = Var(
            doc="Equilibrium constant coefficient 1", units=pyunits.dimensionless
        )
        set_param_from_config(rblock, param="k_eq_coeff", index="1", config=config)

        rblock.k_eq_coeff_2 = Var(
            doc="Equilibrium constant coefficient 2", units=pyunits.K
        )
        set_param_from_config(rblock, param="k_eq_coeff", index="2", config=config)

        rblock.k_eq_coeff_3 = Var(
            doc="Equilibrium constant coefficient 3", units=pyunits.dimensionless
        )
        set_param_from_config(rblock, param="k_eq_coeff", index="3", config=config)

    @staticmethod
    def return_expression(b, rblock, r_idx, T):
        return exp(b.log_k_eq[r_idx]) * ((pyunits.m) ** 3 / pyunits.mol)

    @staticmethod
    def return_log_expression(b, rblock, r_idx, T):
        return b.log_k_eq[r_idx] == (
            rblock.k_eq_coeff_1
            + rblock.k_eq_coeff_2 / T
            + rblock.k_eq_coeff_3 * log(T / pyunits.K)
            + log(1e-3)
        )

    @staticmethod
    def calculate_scaling_factors(b, rblock):
        return 1


# -----------------------------------------------------------------------------
# Configuration dictionary for aqueous MEA solvent
configuration = {
    # Specifying components
    "components": {
        "H2O": {
            "type": Solvent,
            "cp_mol_liq_comp": CpMolSolvent,
            "diffus_phase_comp": {"Liq": DiffusNone},
            "enth_mol_liq_comp": EnthMolSolvent,
            "pressure_sat_comp": PressureSatSolvent,
            "vol_mol_liq_comp": VolMolSolvent,
            "parameter_data": {
                "mw": (0.01802, pyunits.kg / pyunits.mol),
                "cp_mass_liq_comp_coeff": {
                    "1": 4.2107,
                    "2": -1.696e-3,
                    "3": 2.568e-5,
                    "4": -1.095e-7,
                    "5": 3.038e-10,
                },
                "dens_mol_liq_comp_coeff": {
                    "1": (-3.2484e-6, pyunits.g / pyunits.mL / pyunits.K**2),  # [2]
                    "2": (0.00165, pyunits.g / pyunits.mL / pyunits.K),
                    "3": (0.793, pyunits.g / pyunits.mL),
                },
                "dh_vap": 43.99e3,
                "pressure_sat_comp_coeff": {
                    "1": 72.55,
                    "2": -7206.70,
                    "3": -7.1385,
                    "4": 4.05e-6,
                },
                "temperature_crit": (647.13, pyunits.K),
            },
        },
        "MEA": {
            "type": Solvent,
            "cp_mol_liq_comp": CpMolSolvent,
            "diffus_phase_comp": {"Liq": DiffusMEA},
            "enth_mol_liq_comp": EnthMolSolvent,
            "pressure_sat_comp": PressureSatSolvent,
            "vol_mol_liq_comp": VolMolSolvent,
            "parameter_data": {
                "mw": (0.06108, pyunits.kg / pyunits.mol),
                "cp_mass_liq_comp_coeff": {
                    "1": 2.6161,
                    "2": 3.706e-3,
                    "3": 3.787e-6,
                    "4": 0.0,
                    "5": 0.0,
                },
                "dens_mol_liq_comp_coeff": {
                    "1": (-5.35162e-7, pyunits.g / pyunits.mL / pyunits.K**2),  # [2]
                    "2": (-4.51417e-4, pyunits.g / pyunits.mL / pyunits.K),
                    "3": (1.19451, pyunits.g / pyunits.mL),
                },
                "dh_vap": 58000,  # [3]
                "diffus_phase_comp_coeff": {
                    "1": -13.275,
                    "2": -2198.3,
                    "3": -7.8142e-5,
                },
                "pressure_sat_comp_coeff": {
                    "1": 172.78,
                    "2": -13492,
                    "3": -21.914,
                    "4": 1.38e-5,
                },
                "temperature_crit": (614.45, pyunits.K),
            },
        },
        "CO2": {
            "type": Solute,
            "cp_mol_liq_comp": CpMolCO2,
            "diffus_phase_comp": {"Liq": DiffusCO2},
            "enth_mol_liq_comp": EnthMolCO2,
            "henry_component": {
                "Liq": {
                    "method": N2OAnalogy,
                    "type": HenryType.Kpc,
                    "basis": StateIndex.true,
                }
            },
            "vol_mol_liq_comp": VolMolCO2,
            "parameter_data": {
                "mw": (0.04401, pyunits.kg / pyunits.mol),
                "dh_abs_co2": -84000,
                "lwm_coeff": {
                    "1": 1.70981,
                    "2": 0.03972,
                    "3": -4.3e-4,
                    "4": -2.20377,
                },
                "diffus_phase_comp_coeff": {
                    "1": 2.35e-6,
                    "2": 2.9837e-8,
                    "3": -9.7078e-9,
                    "4": -2119,
                    "5": -20.132,
                },
                "vol_mol_liq_comp_coeff": {
                    "a": (10.2074, pyunits.mL / pyunits.mol),  # [2]
                    "b": (-2.2642, pyunits.mL / pyunits.mol),
                    "c": (3.0059, pyunits.mL / pyunits.mol),
                    "d": (207, pyunits.mL / pyunits.mol),
                    "e": (-563.3701, pyunits.mL / pyunits.mol),
                },
            },
        },
        "MEA_+": {
            "type": Cation,
            "charge": +1,
            "diffus_phase_comp": {"Liq": DiffusIons},
            "parameter_data": {
                "diffus_phase_comp_coeff": {"1": -22.64, "2": -1000.0, "3": -0.7}
            },
        },
        "MEACOO_-": {
            "type": Anion,
            "charge": -1,
            "diffus_phase_comp": {"Liq": DiffusIons},
            "parameter_data": {
                "diffus_phase_comp_coeff": {"1": -22.64, "2": -1000.0, "3": -0.7}
            },
        },
        "HCO3_-": {
            "type": Anion,
            "charge": -1,
            "diffus_phase_comp": {"Liq": DiffusNone},
        },
    },
    # Specifying phases
    "phases": {
        "Liq": {
            "type": AqueousPhase,
            "equation_of_state": Ideal,
            "equation_of_state_options": {"property_basis": "apparent"},
            "surf_tens_phase": SurfTens,
            "therm_cond_phase": ThermalCond,
            "visc_d_phase": Viscosity,
            "parameter_data": {
                "surf_tens_H2O_coeff": {
                    "1": 0.18548,
                    "2": 2.717,
                    "3": -3.554,
                    "4": 2.047,
                },
                "surf_tens_MEA_coeff": {"1": 0.09945, "2": 1.067, "3": 0.0, "4": 0.0},
                "surf_tens_CO2_coeff": {
                    "1": -5.987,
                    "2": 3.7699,
                    "3": -0.43164,
                    "4": 0.018155,
                    "5": -0.01207,
                    "6": 0.002119,
                },
                "surf_tens_F_coeff": {
                    "a": 2.4558,
                    "b": -1.5311,
                    "c": 3.4994,
                    "d": -5.6398,
                    "e": 10.2109,
                    "f": 2.3122,
                    "g": 4.5608,
                    "h": -2.3924,
                    "i": 5.3324,
                    "j": -12.0494,
                },
                "visc_d_coeff": {
                    "a": -0.0838,
                    "b": 2.8817,
                    "c": 33.651,
                    "d": 1817.0,
                    "e": 0.00847,
                    "f": 0.0103,
                    "g": -2.3890,
                },
            },
        }
    },
    # Set base units of measurement
    "base_units": {
        "time": pyunits.s,
        "length": pyunits.m,
        "mass": pyunits.kg,
        "amount": pyunits.mol,
        "temperature": pyunits.K,
    },
    # Specifying state definition
    "state_definition": FTPx,
    "state_bounds": {
        "flow_mol": (0, 1, 1000, pyunits.mol / pyunits.s),
        "temperature": (273.15, 298.15, 450, pyunits.K),
        "pressure": (5e4, 101325, 1e6, pyunits.Pa),
    },
    "state_components": StateIndex.apparent,
    "pressure_ref": (101325, pyunits.Pa),
    "temperature_ref": (298.15, pyunits.K),
    "inherent_reactions": {
        "carbamate": {
            "stoichiometry": {
                ("Liq", "MEA"): -2,
                ("Liq", "CO2"): -1,
                ("Liq", "MEA_+"): 1,
                ("Liq", "MEACOO_-"): 1,
            },
            "equilibrium_constant": k_eq,
            "equilibrium_form": log_power_law_equil,
            "concentration_form": ConcentrationForm.molarity,
            "parameter_data": {"k_eq_coeff": {"1": 233.4, "2": -3410, "3": -36.8}},
        },
        "bicarbonate": {
            "stoichiometry": {
                ("Liq", "MEA"): -1,
                ("Liq", "CO2"): -1,
                ("Liq", "H2O"): -1,
                ("Liq", "HCO3_-"): 1,
                ("Liq", "MEA_+"): 1,
            },
            "equilibrium_constant": k_eq,
            "equilibrium_form": log_power_law_equil,
            "concentration_form": ConcentrationForm.molarity,
            "parameter_data": {"k_eq_coeff": {"1": 176.72, "2": -2909, "3": -28.46}},
        },
    },
}
